import os
import tempfile
from typing import Any

from common.secrets import get_secrets

from x8.storage.object_store import ObjectStore

secrets = get_secrets()


class ObjectStoreProvider:
    AMAZON_S3 = "amazon_s3"
    AZURE_BLOB_STORAGE = "azure_blob_storage"
    GOOGLE_CLOUD_STORAGE = "google_cloud_storage"
    FILE_SYSTEM = "file_system"
    MEMORY = "memory"


provider_parameters: dict[str, dict[str, Any]] = {
    ObjectStoreProvider.AMAZON_S3: {
        "region_name": secrets["storage-object-amazon-s3-region-name"],
        "aws_access_key_id": secrets[
            "storage-object-amazon-s3-aws-access-key-id"
        ],
        "aws_secret_access_key": secrets[
            "storage-object-amazon-s3-aws-secret-access-key"
        ],
        "bucket": "x8-test-versioned",
    },
    ObjectStoreProvider.AZURE_BLOB_STORAGE: {
        "connection_string": secrets[
            "storage-object-azure-blob-storage-connection-string"
        ],
        "container": "x8-test-versioned",
    },
    ObjectStoreProvider.GOOGLE_CLOUD_STORAGE: {
        "project": secrets["storage-object-google-cloud-storage-project"],
        "service_account_info": secrets[
            "storage-object-google-cloud-storage-service-account-info"
        ],
        "bucket": "x8-test-versioned",
    },
    ObjectStoreProvider.FILE_SYSTEM: {
        "store_path": os.path.join(tempfile.gettempdir(), "object-store"),
        "folder": "x8-test-versioned",
    },
}


def get_component(provider_type: str, collection: str | None = None):
    parameters = provider_parameters[provider_type]
    component = ObjectStore(
        collection=collection,
        __provider__=dict(
            type=provider_type,
            parameters=parameters,
        ),
    )
    return component



from x8.core import DataModel


class AudioData(DataModel):
    source: str | None = None
    content: bytes | str | None = None
    media_type: str | None = None


class AudioInfo(DataModel):
    duration: float
    """Duration of the audio in seconds."""

    channels: int
    """Number of audio channels."""

    sample_rate: int
    """Sample rate of the audio."""

    bit_rate: int
    """Bit rate of the audio."""

    format: str | None = None
    """Format of the audio."""

    codec: str | None = None
    """Name of the codec used for the audio stream."""



"""
Docker local for container registry.
"""

__all__ = ["DockerLocal"]

from datetime import datetime

import docker
from x8.core import Provider, Response
from x8.core.exceptions import BadRequestError

from .._models import ContainerRegistryItem, ContainerRegistryItemDigest


class DockerLocal(Provider):
    def __init__(
        self,
        **kwargs,
    ):
        """Initialize."""
        super().__init__(**kwargs)

    def push(self, image_name: str) -> Response[ContainerRegistryItem]:
        client = docker.from_env()
        image = client.images.get(image_name)
        client.images.push(image_name)
        result = ContainerRegistryItem(
            image_name=image_name, image_uri=image.tags[0]
        )
        return Response(result=result)

    def pull(
        self,
        image_name: str,
        tag: str | None = None,
    ) -> Response[ContainerRegistryItem]:
        client = docker.from_env()
        if tag is None:
            image = client.images.get(image_name)
        else:
            image = client.images.get(f"{image_name}:{tag}")
        result = ContainerRegistryItem(
            image_name=image_name,
            image_uri=image.tags[0],
        )
        return Response(result=result)

    def tag(
        self,
        image_name: str,
        tag: str,
        digest: str | None = None,
    ) -> Response[ContainerRegistryItemDigest]:
        image = self._get_image(image_name, digest)
        image.tag(image_name, tag=tag)
        result = self._get_digest(image, image.tags[0])
        return Response(result=result)

    def delete(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        client = docker.from_env()

        if digest:
            image_ref = f"{image_name}@{digest}"
            client.images.remove(image=image_ref, force=True)
        elif tag:
            image_ref = f"{image_name}:{tag}"
            client.images.remove(image=image_ref, force=True)
        else:
            images = client.images.list(name=image_name)
            for image in images:
                if any(tag.startswith(image_name) for tag in image.tags):
                    client.images.remove(image.id, force=True)

        return Response(result=None)

    def list_images(
        self,
    ) -> Response[list[ContainerRegistryItem]]:
        client = docker.from_env()
        images = client.images.list()
        image_items = []
        for image in images:
            if image.tags:
                image_items.append(
                    ContainerRegistryItem(
                        image_name=image.tags[0].split(":")[0],
                        image_uri=image.tags[0],
                    )
                )
        result = image_items
        return Response(result=result)

    def get_digests(
        self,
        image_name: str,
    ) -> Response[list[ContainerRegistryItemDigest]]:
        client = docker.from_env()
        images = client.images.list(name=image_name)

        if not images:
            return Response(result=[])

        digests = []
        for image in images:
            for tag in image.tags:
                if tag.startswith(image_name):
                    digests.append(self._get_digest(image, tag=tag))

        return Response(result=digests)

    def _get_image(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ):
        client = docker.from_env()
        if tag is not None:
            image_ref = f"{image_name}:{tag}"
        else:
            image_ref = image_name
        if digest is None:
            image = client.images.get(image_ref)
        else:
            matching_images = [
                img
                for img in client.images.list()
                if f"{image_name}@{digest}"
                in (img.attrs.get("RepoDigests") or [])
            ]
            if not matching_images:
                raise BadRequestError(f"No image found with digest {digest}")
            image = matching_images[0]
        return image

    def _get_digest(self, image, tag: str) -> ContainerRegistryItemDigest:
        return ContainerRegistryItemDigest(
            image_uri=tag,
            digest=image.id,
            image_size_bytes=image.attrs["Size"],
            tags=[tag.split(":")[1] for tag in image.tags],
            upload_time=datetime.fromisoformat(
                image.attrs["Created"].replace("Z", "+00:00")
            ).timestamp(),
        )

    def close(self) -> Response[None]:
        return Response(result=None)



from x8.core.exceptions import ConflictError, NotFoundError
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)

from ._feature import PubSubFeature
from .component import PubSub

__all__ = [
    "MessageBatch",
    "MessageItem",
    "MessageProperties",
    "MessagePullConfig",
    "MessagePutConfig",
    "MessageValueType",
    "PubSub",
    "PubSubFeature",
    "NotFoundError",
    "ConflictError",
    "SubscriptionConfig",
    "SubscriptionInfo",
    "TopicConfig",
    "TopicInfo",
]



from fastapi import HTTPException, status


class UnauthenticatedException(HTTPException):
    def __init__(self, detail: str = "Requires authentication"):
        super().__init__(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=detail,
        )


class UnauthorizedException(HTTPException):
    def __init__(self, detail: str = "Not authorized", **kwargs):
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=detail,
        )


class NotFoundException(HTTPException):
    def __init__(self, detail: str = "Item not found"):
        super().__init__(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=detail,
        )


class ConflictException(HTTPException):
    def __init__(self, detail: str = "Item already exists"):
        super().__init__(
            status_code=status.HTTP_409_CONFLICT,
            detail=detail,
        )



kvs = [
    {
        "id": "id00",
        "value": b"Hello World 1",
    },
    {
        "id": "id01",
        "value": 10,
    },
    {
        "id": "id02",
        "value": "Hello World ",
    },
]

batch_kvs = [
    {
        "id": "id00",
        "value": b"Batch 0",
        "collection": "col1",
    },
    {
        "id": "id01",
        "value": b"Batch 1",
        "collection": "col2",
    },
    {
        "id": "id02",
        "value": b"Batch 2",
        "collection": "col1",
    },
    {
        "id": "id03",
        "value": b"Batch 3",
        "collection": "col2",
    },
]

query_kvs = [
    {
        "id": "q00",
        "value": b"value0",
        "collection": "col1",
    },
    {
        "id": "q01",
        "value": b"value1",
        "collection": "col1",
    },
    {
        "id": "q02",
        "value": b"value2",
        "collection": "col1",
    },
    {
        "id": "q03",
        "value": b"value3",
        "collection": "col1",
    },
    {
        "id": "q04",
        "value": b"value4",
        "collection": "col1",
    },
    {
        "id": "q05",
        "value": b"value5",
        "collection": "col2",
    },
    {
        "id": "q06",
        "value": b"value6",
        "collection": "col2",
    },
    {
        "id": "q07",
        "value": b"value7",
        "collection": "col2",
    },
    {
        "id": "q08",
        "value": b"value8",
        "collection": "col2",
    },
    {
        "id": "q09",
        "value": b"value9",
        "collection": "col2",
    },
    {
        "id": "q10",
        "value": b"value10",
        "collection": "col2",
    },
    {
        "id": "q11",
        "value": b"value11",
        "collection": "col2",
    },
    {
        "id": "q12",
        "value": b"value12",
        "collection": "col2",
    },
    {
        "id": "q13",
        "value": b"value13",
        "collection": "col2",
    },
]



from x8.compute.containerizer import BuildConfig, PrepareConfig
from x8.core import Component, DataModel


class ImageMap(DataModel):
    name: str
    local_image: str | None = None
    source: str | None = None
    component: Component | None = None
    handle: str | None = None
    prepare: PrepareConfig = PrepareConfig()
    build: BuildConfig = BuildConfig()



"""
Google cloud run container deployment.
"""

from __future__ import annotations

__all__ = ["GoogleCloudRun"]

from typing import Any

from google.api_core.exceptions import FailedPrecondition, NotFound
from google.cloud import run_v2

from x8._common.google_provider import GoogleProvider
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, OperationParser, Response, RunContext
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import Expression

from .._helper import amap_images, map_images, requires_container_registry
from .._models import (
    Container,
    EnvVar,
    GPUResource,
    GRPCAction,
    HTTPGetAction,
    HTTPHeader,
    Ingress,
    Port,
    Probe,
    ProbeSet,
    ResourceLimits,
    ResourceRequirements,
    Revision,
    Scale,
    ServiceDefinition,
    ServiceItem,
    ServiceOverlay,
    TCPSocketAction,
    TrafficAllocation,
    Volume,
    VolumeMount,
)
from ._base import BaseContainerDeploymentProvider


class GoogleCloudRun(GoogleProvider, BaseContainerDeploymentProvider):
    project: str | None
    location: str
    service_name: str | None
    nparams: dict[str, Any]

    _credentials: Any
    _services_client: run_v2.ServicesClient
    _aservices_client: run_v2.ServicesAsyncClient
    _revisions_client: run_v2.RevisionsClient
    _arevisions_client: run_v2.RevisionsAsyncClient
    _init: bool
    _ainit: bool
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _cached_project: str | None

    def __init__(
        self,
        project: str | None = None,
        location: str = "us-central1",
        service_name: str | None = None,
        service: ServiceDefinition | None = None,
        service_override: ServiceOverlay | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        allow_unauthenticated_access: bool | None = True,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            project:
                Google project id.
            location:
                Google cloud location.
            service_name:
                Cloud run service name.
            service:
                Service definition to deploy.
            service_override:
                Service override definition to apply.
            containerizer:
                Containerizer to use for building the container.
            container_registry:
                Container registry to use for pushing the container.
            service_account_info:
                Google service account info with serialized credentials.
            service_account_file:
                Google service account file with credentials.
            access_token:
                Google access token.
            nparams:
                Native params to google client.
        """
        self.project = project
        self.location = location
        self.service_name = service_name
        self.service = service
        self.overlay = service_override
        self.containerizer = containerizer
        self.container_registry = container_registry
        self.allow_unauthenticated_access = allow_unauthenticated_access
        self.nparams = nparams

        self._credentials = None
        self._init = False
        self._ainit = False
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()
        self._cached_project = None

        GoogleProvider.__init__(
            self,
            service_account_info=service_account_info,
            service_account_file=service_account_file,
            access_token=access_token,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        self._credentials = self._get_credentials()
        self._services_client = run_v2.ServicesClient(
            credentials=self._credentials
        )
        self._revisions_client = run_v2.RevisionsClient(
            credentials=self._credentials
        )
        self._init = True

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._ainit:
            return

        self._credentials = self._get_credentials()
        self._aservices_client = run_v2.ServicesAsyncClient(
            credentials=self._credentials
        )
        self._arevisions_client = run_v2.RevisionsAsyncClient(
            credentials=self._credentials
        )
        self._ainit = True

    def _ensure_container_registry(
        self, service: ServiceDefinition
    ) -> ContainerRegistry | None:
        container_registry = self._get_container_registry()
        if container_registry:
            return container_registry
        if not requires_container_registry(service):
            return None

        from x8.compute.container_registry.providers.google_artifact_registry import (  # noqa
            GoogleArtifactRegistry,
        )

        service_name = self.service_name or service.name
        gcr_name = f"{service_name}-gcr"
        gcr = GoogleArtifactRegistry(
            project=self.project,
            location=self.location,
            name=gcr_name,
            service_account_info=self.service_account_info,
            service_account_file=self.service_account_file,
            access_token=self.access_token,
        )
        gcr.create_resource(format="DOCKER", mode="STANDARD")
        print(f"Created container registry: {gcr_name}")
        return ContainerRegistry(__provider__=gcr)

    def _check_delete_container_registry(self, name: str) -> None:
        container_registry = self._get_container_registry()
        if container_registry:
            return None

        from x8.compute.container_registry.providers.google_artifact_registry import (  # noqa
            GoogleArtifactRegistry,
        )

        gcr_name = f"{name}-gcr"
        gcr = GoogleArtifactRegistry(
            project=self.project,
            location=self.location,
            name=gcr_name,
            service_account_info=self.service_account_info,
            service_account_file=self.service_account_file,
            access_token=self.access_token,
        )
        try:
            gcr.get_resource()
            gcr.delete_resource()
            print(f"Deleted container registry: {gcr_name}")
        except NotFoundError:
            pass

    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        self.__setup__()
        client = self._services_client
        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        project = self._get_project()
        service_name, parent_path, service_path = self._get_service_paths(
            project, service_def.name
        )
        try:
            existing_service = client.get_service(name=service_path)
            if where_exists is False:
                raise PreconditionFailedError(
                    f"Service {service_name} already exists."
                )
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            updated_service = self._op_converter.update_service(
                existing_service,
                service_def,
                images,
            )
            operation = client.update_service(service=updated_service)
            response = operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service updated successfully: {response.uri}")
        except NotFound:
            if where_exists is True:
                raise PreconditionFailedError(
                    f"Service {service_name} not found."
                )
            images = map_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            gcr_service = self._op_converter.convert_service(
                service_def, images
            )
            operation = client.create_service(
                parent=parent_path,
                service=gcr_service,
                service_id=service_name,
            )
            response = operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service created successfully: {response.uri}")
        if not service_def.ingress or service_def.ingress.external:
            self._allow_unauthenticated_access(project, service_name)
        return Response(result=result)

    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        await self.__asetup__()
        client = self._aservices_client
        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        project = self._get_project()
        service_name, parent_path, service_path = self._get_service_paths(
            project, service_def.name
        )
        try:
            existing_service = await client.get_service(name=service_path)
            if where_exists is False:
                raise PreconditionFailedError(
                    f"Service {service_name} already exists."
                )
            images = await amap_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            updated_service = self._op_converter.update_service(
                existing_service,
                service_def,
                images,
            )
            operation = await client.update_service(service=updated_service)
            response = await operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service updated successfully: {response.uri}")
        except NotFound:
            if where_exists is True:
                raise PreconditionFailedError(
                    f"Service {service_name} not found."
                )
            images = await amap_images(
                containers=service_def.containers,
                images=service_def.images,
                containerizer=self._get_containerizer(),
                container_registry=self._ensure_container_registry(
                    service_def
                ),
                run_context=run_context,
            )
            gcr_service = self._op_converter.convert_service(
                service_def, images
            )
            operation = await client.create_service(
                parent=parent_path,
                service=gcr_service,
                service_id=service_name,
            )
            response = await operation.result()
            result = self._result_converter.convert_service_item(response)
            print(f"Service created successfully: {response.uri}")
        if not service_def.ingress or service_def.ingress.external:
            self._allow_unauthenticated_access(project, service_name)
        return Response(result=result)

    def get_service(self, name: str) -> Response[ServiceItem]:
        self.__setup__()
        client = self._services_client
        project = self._get_project()
        try:
            response = client.get_service(
                name=self._get_service_path(project, name)
            )
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    async def aget_service(self, name: str) -> Response[ServiceItem]:
        await self.__asetup__()
        client = self._aservices_client
        project = self._get_project()
        try:
            response = await client.get_service(
                name=self._get_service_path(project, name)
            )
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    def delete_service(self, name: str) -> Response[None]:
        self.__setup__()
        client = self._services_client
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        try:
            client.delete_service(name=service_path)
            print(f"Service {name} deleted successfully.")
            self._check_delete_container_registry(name)
            return Response(result=None)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    async def adelete_service(self, name: str) -> Response[None]:
        await self.__asetup__()
        client = self._aservices_client
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        try:
            await client.delete_service(name=service_path)
            print(f"Service {name} deleted successfully.")
            self._check_delete_container_registry(name)
            return Response(result=None)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    def list_services(self) -> Response[list[ServiceItem]]:
        self.__setup__()
        project = self._get_project()
        parent_path = f"projects/{project}/locations/{self.location}"
        request = run_v2.ListServicesRequest(
            parent=parent_path,
            page_size=1000,
            show_deleted=False,
        )
        services = self._services_client.list_services(request=request)
        result = self._result_converter.convert_services(
            services=list(services)
        )
        return Response(result=result)

    async def alist_services(self) -> Response[list[ServiceItem]]:
        await self.__asetup__()
        project = self._get_project()
        parent_path = f"projects/{project}/locations/{self.location}"
        request = run_v2.ListServicesRequest(
            parent=parent_path,
            page_size=1000,
            show_deleted=False,
        )
        services = await self._aservices_client.list_services(request=request)
        result = self._result_converter.convert_services(
            services=[service async for service in services]
        )
        return Response(result=result)

    def list_revisions(self, name: str) -> Response[list[Revision]]:
        self.__setup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        request = run_v2.ListRevisionsRequest(
            parent=service_path,
            page_size=1000,
            show_deleted=False,
        )
        try:
            service = self._services_client.get_service(name=service_path)
            revisions = self._revisions_client.list_revisions(request=request)
            result = self._result_converter.convert_revisions(
                revisions=list(revisions),
                service=service,
            )
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    async def alist_revisions(self, name: str) -> Response[list[Revision]]:
        await self.__asetup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        request = run_v2.ListRevisionsRequest(
            parent=service_path,
            page_size=1000,
            show_deleted=False,
        )
        try:
            service = await self._aservices_client.get_service(
                name=service_path
            )
            revisions = await self._arevisions_client.list_revisions(
                request=request
            )
            result = self._result_converter.convert_revisions(
                revisions=[rev async for rev in revisions],
                service=service,
            )
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    def get_revision(self, name: str, revision: str) -> Response[Revision]:
        self.__setup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        revision_path = f"{service_path}/revisions/{revision}"
        try:
            service = self._services_client.get_service(name=service_path)
            revision = self._revisions_client.get_revision(name=revision_path)
            result = self._result_converter.convert_revision(
                revision=revision,
                service=service,
            )
            return Response(result=result)
        except NotFound:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )

    async def aget_revision(
        self, name: str, revision: str
    ) -> Response[Revision]:
        await self.__asetup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        revision_path = f"{service_path}/revisions/{revision}"
        try:
            service = await self._aservices_client.get_service(
                name=service_path
            )
            revision = await self._arevisions_client.get_revision(
                name=revision_path
            )
            result = self._result_converter.convert_revision(
                revision=revision,
                service=service,
            )
            return Response(result=result)
        except NotFound:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )

    def delete_revision(self, name: str, revision: str) -> Response[None]:
        self.__setup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        revision_path = f"{service_path}/revisions/{revision}"
        try:
            self._revisions_client.delete_revision(name=revision_path)
            return Response(result=None)
        except NotFound:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )
        except FailedPrecondition:
            raise PreconditionFailedError(
                f"Revision {revision} of service {name} cannot be deleted."
            )

    async def adelete_revision(
        self, name: str, revision: str
    ) -> Response[None]:
        await self.__asetup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        revision_path = f"{service_path}/revisions/{revision}"
        try:
            await self._arevisions_client.delete_revision(name=revision_path)
            return Response(result=None)
        except NotFound:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )
        except FailedPrecondition:
            raise PreconditionFailedError(
                f"Revision {revision} of service {name} cannot be deleted."
            )

    def update_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        self.__setup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        try:
            service = self._services_client.get_service(name=service_path)
            traffic_targets = self._op_converter.convert_traffic(traffic)
            service.traffic = traffic_targets
            operation = self._services_client.update_service(service=service)
            response = operation.result()
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    async def aupdate_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        await self.__asetup__()
        project = self._get_project()
        service_path = self._get_service_path(project, name)
        try:
            service = await self._aservices_client.get_service(
                name=service_path
            )
            traffic_targets = self._op_converter.convert_traffic(traffic)
            service.traffic = traffic_targets
            operation = await self._aservices_client.update_service(
                service=service
            )
            response = await operation.result()
            result = self._result_converter.convert_service_item(response)
            return Response(result=result)
        except NotFound:
            raise NotFoundError(f"Service {name} not found.")

    def close(self) -> None:
        self._init = False

    async def aclose(self) -> None:
        self._ainit = False

    def _get_project(self) -> str:
        if self.project:
            return self.project
        if self._cached_project:
            return self._cached_project
        self._cached_project = self._get_default_project()
        if not self._cached_project:
            raise BadRequestError("Project is required.")
        return self._cached_project

    def _get_service_path(self, project: str, service_name: str) -> str:
        parent_path = f"projects/{project}/locations/{self.location}"
        service_path = f"{parent_path}/services/{service_name}"
        return service_path

    def _get_service_paths(
        self,
        project: str,
        service_name: str | None,
    ) -> tuple[str, str, str]:
        service_name = self.service_name or service_name
        if not service_name:
            raise BadRequestError("Service name is required.")
        parent_path = f"projects/{project}/locations/{self.location}"
        service_path = f"{parent_path}/services/{service_name}"
        return service_name, parent_path, service_path

    def _allow_unauthenticated_access(self, project: str, service_name: str):
        from googleapiclient.discovery import build

        client = build("run", "v1")
        parent = f"projects/{project}/locations/{self.location}"
        service_path = f"{parent}/services/{service_name}"
        request = (
            client.projects()
            .locations()
            .services()
            .getIamPolicy(resource=service_path)
        )
        policy = request.execute()
        binding = {
            "role": "roles/run.invoker",
            "members": ["allUsers"],
        }
        policy.setdefault("bindings", []).append(binding)
        set_policy_request = (
            client.projects()
            .locations()
            .services()
            .setIamPolicy(
                resource=service_path,
                body={"policy": policy},
            )
        )
        set_policy_request.execute()
        print("Allowed unauthenticated access")


class OperationConverter:
    def update_service(
        self,
        existing_service: run_v2.Service,
        service: ServiceDefinition,
        images: list[str],
    ) -> run_v2.Service:
        gcr_service = self.convert_service(
            service=service,
            images=images,
        )
        existing_service.template.containers = gcr_service.template.containers
        existing_service.template.scaling = gcr_service.template.scaling
        existing_service.template.volumes = gcr_service.template.volumes
        existing_service.scaling = gcr_service.scaling
        existing_service.ingress = gcr_service.ingress
        return existing_service

    def convert_service(
        self,
        service: ServiceDefinition,
        images: list[str],
    ) -> run_v2.Service:
        containers: list[run_v2.Container] = []
        for container, image in zip(service.containers, images):
            containers.append(
                self._convert_container(
                    container=container,
                    image=image,
                    service=service,
                )
            )
        return run_v2.Service(
            template=run_v2.RevisionTemplate(
                containers=containers,
                scaling=self._convert_revision_scaling(
                    scale=service.scale,
                ),
                volumes=self._convert_volumes(service.volumes),
            ),
            scaling=self._convert_service_scaling(
                scale=service.scale,
            ),
            ingress=self._convert_ingress(
                ingress=service.ingress,
            ),
            traffic=self.convert_traffic(
                traffic=service.traffic,
            ),
        )

    def _convert_ingress(
        self, ingress: Ingress | None
    ) -> run_v2.IngressTraffic:
        if ingress is None:
            return run_v2.IngressTraffic(
                run_v2.IngressTraffic.INGRESS_TRAFFIC_ALL
            )
        if ingress.external:
            return run_v2.IngressTraffic(
                run_v2.IngressTraffic.INGRESS_TRAFFIC_ALL
            )
        return run_v2.IngressTraffic(
            run_v2.IngressTraffic.INGRESS_TRAFFIC_INTERNAL_ONLY
        )

    def convert_traffic(
        self, traffic: list[TrafficAllocation] | None
    ) -> list[run_v2.TrafficTarget]:
        if traffic is None:
            return [
                run_v2.TrafficTarget(
                    percent=100,
                    type_=run_v2.TrafficTargetAllocationType(
                        run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST  # noqa
                    ),
                )
            ]
        result: list[run_v2.TrafficTarget] = []
        for item in traffic:
            tt = run_v2.TrafficTarget(
                revision=item.revision,
                percent=int(item.percent),
            )
            if item.latest_revision:
                tt.type_ = run_v2.TrafficTargetAllocationType(
                    run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST  # noqa
                )
            else:
                tt.type_ = run_v2.TrafficTargetAllocationType(
                    run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_REVISION  # noqa
                )
            if item.tag:
                tt.tag = item.tag
            result.append(tt)
        return result

    def _convert_volumes(
        self,
        volumes: list[Volume],
    ) -> list[run_v2.Volume]:
        result: list[run_v2.Volume] = []
        for volume in volumes:
            if volume.type == "emptyDir":
                result.append(
                    run_v2.Volume(
                        name=volume.name,
                        empty_dir=run_v2.EmptyDirVolumeSource(
                            size_limit=volume.size_limit,
                        ),
                    )
                )
        return result

    def _convert_service_scaling(
        self, scale: Scale | None
    ) -> run_v2.ServiceScaling:
        if scale is None:
            return run_v2.ServiceScaling(
                scaling_mode=run_v2.ServiceScaling.ScalingMode.AUTOMATIC,
                min_instance_count=0,
            )
        if scale.mode == "manual":
            return run_v2.ServiceScaling(
                scaling_mode=run_v2.ServiceScaling.ScalingMode.MANUAL,
                manual_instance_count=scale.replicas or 1,
            )
        return run_v2.ServiceScaling(
            scaling_mode=run_v2.ServiceScaling.ScalingMode.AUTOMATIC,
            min_instance_count=scale.min_replicas or 1,
        )

    def _convert_revision_scaling(
        self,
        scale: Scale | None,
    ) -> run_v2.RevisionScaling:
        if scale is None:
            return run_v2.RevisionScaling(
                min_instance_count=0,
                max_instance_count=10,
            )
        return run_v2.RevisionScaling(
            min_instance_count=scale.min_replicas or 0,
            max_instance_count=scale.max_replicas or 10,
        )

    def _convert_container(
        self,
        container: Container,
        image: str,
        service: ServiceDefinition,
    ) -> run_v2.Container:
        args: dict[str, Any] = {}
        if container.ports:
            args["ports"] = self._convert_ports(container.ports)
        if container.resources:
            args["resources"] = self._convert_resource_requirements(
                container.resources
            )
        if container.volume_mounts:
            args["volume_mounts"] = self._convert_volume_mounts(
                container.volume_mounts
            )
        if container.probes:
            if container.probes.liveness_probe:
                args["liveness_probe"] = self._convert_probe(
                    container.probes.liveness_probe
                )
            if container.probes.startup_probe:
                args["startup_probe"] = self._convert_probe(
                    container.probes.startup_probe
                )
        depends_on = self._convert_depends_on(
            container_name=container.name,
            service=service,
        )
        if depends_on:
            args["depends_on"] = depends_on
        return run_v2.Container(
            name=container.name,
            image=image,
            command=container.command or [],
            args=container.args or [],
            working_dir=container.working_dir,
            env=self._convert_env(container.env),
            **args,
        )

    def _convert_env(self, env: list[EnvVar]) -> list[run_v2.EnvVar]:
        return [
            run_v2.EnvVar(name=var.name, value=str(var.value)) for var in env
        ]

    def _convert_ports(
        self,
        ports: list[Port],
    ) -> list[run_v2.ContainerPort]:
        return [
            run_v2.ContainerPort(
                container_port=port.container_port,
                name=port.name,
            )
            for port in ports
        ]

    def _convert_resource_requirements(
        self,
        resources: ResourceRequirements,
    ) -> run_v2.ResourceRequirements | None:
        cpu = "1"
        memory = "512Mi"
        limits = {
            "cpu": cpu,
            "memory": memory,
        }
        if resources.limits:
            if resources.limits.cpu:
                cpu = str(resources.limits.cpu)
                limits["cpu"] = cpu
            if resources.limits.memory:
                memory = f"{str(resources.limits.memory)}Mi"
                limits["memory"] = memory
            if resources.limits.gpu and resources.limits.gpu.count:
                limits["nvidia.com/gpu"] = str(resources.limits.gpu.count)

        return run_v2.ResourceRequirements(
            limits=limits,
            cpu_idle=resources.cpu_idle,
            startup_cpu_boost=resources.cpu_boost,
        )

    def _convert_volume_mounts(
        self,
        volume_mounts: list[VolumeMount],
    ) -> list[run_v2.VolumeMount] | None:
        return [
            run_v2.VolumeMount(
                name=mount.name,
                mount_path=mount.mount_path,
            )
            for mount in volume_mounts
        ]

    def _convert_probe(self, probe: Probe | None) -> run_v2.Probe | None:
        if not probe:
            return None
        args: dict = {}
        if probe.http_get:
            args["http_get"] = self._convert_http_get_action(probe.http_get)
        elif probe.grpc:
            args["grpc"] = self._convert_grpc_action(probe.grpc)
        elif probe.tcp_socket:
            args["tcp_socket"] = self._convert_tcp_socket_action(
                probe.tcp_socket
            )
        return run_v2.Probe(
            initial_delay_seconds=probe.initial_delay_seconds,
            timeout_seconds=probe.timeout_seconds,
            period_seconds=probe.period_seconds,
            failure_threshold=probe.failure_threshold,
            **args,
        )

    def _convert_http_get_action(
        self, action: HTTPGetAction
    ) -> run_v2.HTTPGetAction:
        headers = [
            run_v2.HTTPHeader(name=header.name, value=header.value)
            for header in action.http_headers
        ]
        return run_v2.HTTPGetAction(
            path=action.path,
            port=action.port,
            http_headers=headers,
        )

    def _convert_grpc_action(self, action: GRPCAction) -> run_v2.GRPCAction:
        return run_v2.GRPCAction(
            port=action.port,
            service=action.service,
        )

    def _convert_tcp_socket_action(
        self, action: TCPSocketAction
    ) -> run_v2.TCPSocketAction:
        return run_v2.TCPSocketAction(
            port=action.port,
        )

    def _convert_depends_on(
        self,
        container_name: str,
        service: ServiceDefinition,
    ) -> list[str]:
        init_containers: list[str] = []
        for container in service.containers:
            if container.type == "init":
                init_containers.append(container.name)
        if container_name in init_containers:
            return []
        return init_containers


class ResultConverter:
    def convert_service_item(
        self,
        service: run_v2.Service,
    ) -> ServiceItem:
        return ServiceItem(
            name=service.name.split("/")[-1],
            uri=service.uri,
            service=self.convert_service(service),
        )

    def convert_service(self, service: run_v2.Service) -> ServiceDefinition:
        return ServiceDefinition(
            name=service.name.split("/")[-1],
            containers=[
                self._convert_container(container)
                for container in service.template.containers
            ],
            ingress=self._convert_ingress(service.ingress),
            traffic=self._convert_traffic(
                list(service.traffic), service.latest_ready_revision
            ),
            volumes=self._convert_volumes(list(service.template.volumes)),
            scale=self._convert_scale(
                service.scaling, service.template.scaling
            ),
            latest_ready_revision=service.latest_ready_revision.split("/")[-1],
            latest_created_revision=service.latest_created_revision.split("/")[
                -1
            ],
        )

    def _convert_container(
        self,
        container: run_v2.Container,
    ) -> Container:
        return Container(
            name=container.name,
            image=container.image,
            command=(list(container.command) if container.command else []),
            args=list(container.args) if container.args else [],
            working_dir=container.working_dir,
            env=[
                EnvVar(name=env.name, value=env.value) for env in container.env
            ],
            ports=[
                Port(container_port=port.container_port, name=port.name)
                for port in container.ports
            ],
            resources=self._convert_resource_requirements(container.resources),
            volume_mounts=[
                VolumeMount(name=mount.name, mount_path=mount.mount_path)
                for mount in container.volume_mounts
            ],
            probes=ProbeSet(
                liveness_probe=self._convert_probe(container.liveness_probe),
                startup_probe=self._convert_probe(container.startup_probe),
            ),
        )

    def _convert_resource_requirements(
        self,
        resources: run_v2.ResourceRequirements | None,
    ) -> ResourceRequirements | None:
        if not resources:
            return None

        # In some protobuf maps, .limits behaves like a dict already;
        # normalize it.
        limits = dict(resources.limits)

        # --- CPU ---
        cpu_raw = limits.get("cpu")
        cpu_cores: float | None = None
        if cpu_raw is not None:
            if isinstance(cpu_raw, (int, float)):
                cpu_cores = float(cpu_raw)
            else:
                s = str(cpu_raw).strip()
                if s.endswith("m"):
                    # millicpu -> cores
                    num = s[:-1] or "0"
                    cpu_cores = float(num) / 1000.0
                else:
                    cpu_cores = float(s)  # e.g. "1", "0.5"

        # --- Memory (normalize to MiB integer) ---
        mem_raw = limits.get("memory")
        memory_mb: int | None = None
        if mem_raw is not None:
            ms = str(mem_raw).strip()
            try:
                if ms.endswith("Mi"):
                    memory_mb = int(float(ms[:-2]))
                elif ms.endswith("Gi"):
                    memory_mb = int(float(ms[:-2]) * 1024)
                elif ms.endswith("Ti"):
                    memory_mb = int(float(ms[:-2]) * 1024 * 1024)
                elif ms.isdigit():
                    # Assume MiB if bare number sneaks through
                    memory_mb = int(ms)
            except ValueError:
                # Leave as None if unparsable; better than crashing
                memory_mb = None

        # --- GPU (if present) ---
        gpu_count = None
        if "nvidia.com/gpu" in limits:
            try:
                gpu_count = int(limits.get("nvidia.com/gpu", 0))
            except (TypeError, ValueError):
                gpu_count = None

        return ResourceRequirements(
            limits=ResourceLimits(
                cpu=cpu_cores,
                memory=memory_mb,
                gpu=(
                    GPUResource(count=gpu_count)
                    if (gpu_count and gpu_count > 0)
                    else None
                ),
            ),
            cpu_idle=resources.cpu_idle,
            cpu_boost=resources.startup_cpu_boost,
        )

    def _convert_ingress(
        self,
        ingress: run_v2.IngressTraffic,
    ) -> Ingress:
        if ingress == run_v2.IngressTraffic.INGRESS_TRAFFIC_ALL:
            return Ingress(external=True)
        return Ingress(external=False)

    def _convert_traffic(
        self,
        traffic: list[run_v2.TrafficTarget],
        latest_revision_name: str,
    ) -> list[TrafficAllocation]:
        result: list[TrafficAllocation] = []
        for item in traffic:
            latest_revision = item.type_ == run_v2.TrafficTargetAllocationType(
                run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST  # noqa
            )
            if not item.revision and latest_revision:
                revision = latest_revision_name
            else:
                revision = item.revision
            result.append(
                TrafficAllocation(
                    revision=revision,
                    percent=item.percent,
                    latest_revision=latest_revision,
                    tag=item.tag,
                )
            )
        return result

    def _convert_volumes(
        self,
        volumes: list[run_v2.Volume],
    ) -> list[Volume]:
        result: list[Volume] = []
        for volume in volumes:
            if volume.empty_dir:
                result.append(
                    Volume(
                        name=volume.name,
                        type="emptyDir",
                        size_limit=volume.empty_dir.size_limit,
                    )
                )
        return result

    def _convert_scale(
        self,
        service_scaling: run_v2.ServiceScaling,
        revision_scaling: run_v2.RevisionScaling,
    ) -> Scale:
        if (
            service_scaling.scaling_mode
            == run_v2.ServiceScaling.ScalingMode.MANUAL
        ):
            return Scale(
                mode="manual",
                replicas=service_scaling.manual_instance_count or 1,
            )
        return Scale(
            mode="auto",
            min_replicas=revision_scaling.min_instance_count or 0,
            max_replicas=revision_scaling.max_instance_count or 10,
        )

    def _convert_probe(
        self,
        probe: run_v2.Probe | None,
    ) -> Probe | None:
        if not probe:
            return None
        args: dict = {}
        if probe.http_get:
            args["http_get"] = self._convert_http_get_action(probe.http_get)
        elif probe.grpc:
            args["grpc"] = self._convert_grpc_action(probe.grpc)
        elif probe.tcp_socket:
            args["tcp_socket"] = self._convert_tcp_socket_action(
                probe.tcp_socket
            )
        return Probe(
            initial_delay_seconds=probe.initial_delay_seconds,
            timeout_seconds=probe.timeout_seconds,
            period_seconds=probe.period_seconds,
            failure_threshold=probe.failure_threshold,
            **args,
        )

    def _convert_http_get_action(
        self, action: run_v2.HTTPGetAction
    ) -> HTTPGetAction:
        headers = [
            HTTPHeader(name=header.name, value=header.value)
            for header in action.http_headers
        ]
        return HTTPGetAction(
            path=action.path,
            port=action.port,
            http_headers=headers,
        )

    def _convert_grpc_action(self, action: run_v2.GRPCAction) -> GRPCAction:
        return GRPCAction(
            port=action.port,
            service=action.service,
        )

    def _convert_tcp_socket_action(
        self, action: run_v2.TCPSocketAction
    ) -> TCPSocketAction:
        return TCPSocketAction(
            port=action.port,
        )

    def convert_revisions(
        self,
        revisions: list[run_v2.Revision],
        service: run_v2.Service,
    ) -> list[Revision]:
        result: list[Revision] = []
        for rev in revisions:
            result.append(self.convert_revision(revision=rev, service=service))
        return result

    def convert_revision(
        self,
        revision: run_v2.Revision,
        service: run_v2.Service,
    ) -> Revision:
        traffic_map = self._get_traffic_map(service)
        return Revision(
            name=revision.name.split("/")[-1],
            traffic=traffic_map.get(revision.name.split("/")[-1], 0),
            created_time=(
                revision.create_time.timestamp()
                if revision.create_time
                else None
            ),  # noqa
            active=self._is_revision_active(revision),
            containers=[
                self._convert_container(container)
                for container in revision.containers
            ],
            volumes=self._convert_volumes(list(revision.volumes)),
        )

    def convert_services(
        self, services: list[run_v2.Service]
    ) -> list[ServiceItem]:
        result: list[ServiceItem] = []
        for service in services:
            result.append(self.convert_service_item(service))
        return result

    def _get_traffic_map(self, service: run_v2.Service) -> dict[str, int]:
        result: dict = {}
        for t in service.traffic:
            if t.type_ == run_v2.TrafficTargetAllocationType(
                run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST  # noqa
            ):
                result[service.latest_ready_revision.split("/")[-1]] = 100
            else:
                result[t.revision] = t.percent
        return result

    def _is_revision_active(self, revision: run_v2.Revision) -> bool:
        if not revision.conditions:
            return False

        for condition in revision.conditions:
            if condition.type_ == "Ready":
                return (
                    condition.state
                    == run_v2.Condition.State.CONDITION_SUCCEEDED
                )
        return False



__all__ = ["AzureKubernetesService", "AzureKubernetesServiceResource"]

from typing import Any

from x8._common.azure_provider import AzureProvider
from x8.compute._common._models import ImageMap
from x8.compute.container_registry.component import ContainerRegistry
from x8.compute.containerizer.component import Containerizer
from x8.core import DataModel, Response
from x8.core.exceptions import BadRequestError, NotFoundError

from .._models import ManifestsType
from ._base import BaseKubernetes


class AzureKubernetesServiceResource(DataModel):
    id: str | None
    name: str | None
    location: str
    kubernetes_version: str | None
    fqdn: str | None
    dns_prefix: str | None


class AzureKubernetesService(AzureProvider, BaseKubernetes):
    resource_group: str | None
    name: str | None

    def __init__(
        self,
        resource_group: str | None = None,
        name: str | None = None,
        kubeconfig: str | dict[str, Any] | None = None,
        context: str | None = None,
        manifests: ManifestsType = None,
        overlays: ManifestsType = None,
        namespace: str | None = None,
        images: list[ImageMap] | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        self.resource_group = resource_group
        self.name = name
        BaseKubernetes.__init__(
            self,
            kubeconfig=kubeconfig,
            context=context,
            manifests=manifests,
            overlays=overlays,
            namespace=namespace,
            images=images,
            containerizer=containerizer,
            container_registry=container_registry,
            **kwargs,
        )
        AzureProvider.__init__(
            self,
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            **kwargs,
        )

    def create_resource(
        self,
        resource_group: str | None = None,
        name: str | None = None,
        location: str = "westus2",
        subscription_id: str | None = None,
        kubernetes_version: str | None = None,
        dns_prefix: str | None = None,
        node_count: int = 1,
        vm_size: str = "Standard_DS2_v2",
        enable_rbac: bool = True,
        network_plugin: str | None = None,
        tags: dict[str, str] | None = None,
        sku_tier: str | None = None,
        enable_managed_identity: bool = True,
        auto_upgrade_channel: str | None = None,
        acr_name: str | None = None,
        acr_resource_group: str | None = None,
    ) -> Response[AzureKubernetesServiceResource]:
        resource_group = resource_group or self.resource_group
        name = name or self.name
        if not name:
            raise BadRequestError("Cluster name is required.")

        if not resource_group:
            raise BadRequestError("Resource group is required.")

        # Ensure RG exists
        self._ensure_resource_group(
            resource_group=resource_group,
            location=location,
            subscription_id=subscription_id,
        )

        from azure.mgmt.containerservice import ContainerServiceClient
        from azure.mgmt.containerservice.models import (
            ContainerServiceNetworkProfile,
            ManagedCluster,
            ManagedClusterAgentPoolProfile,
            ManagedClusterAutoUpgradeProfile,
            ManagedClusterIdentity,
            ManagedClusterSKU,
        )

        client = ContainerServiceClient(
            credential=self._get_credential(),
            subscription_id=(
                subscription_id or self._get_default_subscription_id()
            ),
        )
        agent_pools = [
            ManagedClusterAgentPoolProfile(
                name="nodepool1",
                count=node_count,
                vm_size=vm_size,
                mode="System",
                type="VirtualMachineScaleSets",
                os_type="Linux",
            )
        ]

        sku = None
        if sku_tier:
            sku = ManagedClusterSKU(name="Base", tier=sku_tier)

        net_profile = None
        if network_plugin:
            net_profile = ContainerServiceNetworkProfile(
                network_plugin=network_plugin
            )
        mc = ManagedCluster(
            location=location,
            dns_prefix=dns_prefix or f"{name}-dns",
            kubernetes_version=kubernetes_version,
            enable_rbac=enable_rbac,
            identity=(
                ManagedClusterIdentity(type="SystemAssigned")
                if enable_managed_identity
                else None
            ),
            agent_pool_profiles=agent_pools,
            sku=sku,
            network_profile=net_profile,
            tags=tags,
            auto_upgrade_profile=(
                ManagedClusterAutoUpgradeProfile(
                    upgrade_channel=auto_upgrade_channel
                )
                if auto_upgrade_channel
                else None
            ),
        )

        poller = client.managed_clusters.begin_create_or_update(
            resource_group_name=resource_group,
            resource_name=name,
            parameters=mc,
        )
        result = poller.result()

        # Optionally attach ACR (grant AcrPull to kubelet identity)
        if acr_name:
            try:
                self._attach_acr_to_cluster(
                    subscription_id=str(
                        subscription_id or self._get_default_subscription_id()
                    ),
                    resource_group=resource_group,
                    cluster_name=name,
                    acr_name=acr_name,
                    acr_resource_group=acr_resource_group or resource_group,
                )
            except Exception as e:
                # Surface a concise error but do not fail cluster creation
                print(f"Warning: failed to attach ACR '{acr_name}': {e}")

        resource = AzureKubernetesServiceResource(
            id=result.id,
            name=result.name,
            location=result.location,
            kubernetes_version=result.kubernetes_version,
            fqdn=result.fqdn or result.private_fqdn,
            dns_prefix=result.dns_prefix,
        )
        return Response(result=resource)

    def _attach_acr_to_cluster(
        self,
        subscription_id: str,
        resource_group: str,
        cluster_name: str,
        acr_name: str,
        acr_resource_group: str,
    ) -> None:
        """
        Grant AcrPull role on the ACR to the AKS kubelet identity if available,
        otherwise to the cluster principal. Requires managed identity or SP.
        """
        import uuid

        from azure.mgmt.authorization import AuthorizationManagementClient
        from azure.mgmt.authorization.models import (
            RoleAssignmentCreateParameters,
        )
        from azure.mgmt.containerregistry import (
            ContainerRegistryManagementClient,
        )
        from azure.mgmt.containerservice import ContainerServiceClient

        # Get cluster to obtain identities
        aks_client = ContainerServiceClient(
            credential=self._get_credential(), subscription_id=subscription_id
        )
        mc = aks_client.managed_clusters.get(resource_group, cluster_name)

        # Get ACR resource id
        acr_client = ContainerRegistryManagementClient(
            credential=self._get_credential(), subscription_id=subscription_id
        )
        registry = acr_client.registries.get(
            resource_group_name=acr_resource_group,
            registry_name=acr_name,
        )
        scope = registry.id

        # Determine principal object id to grant
        principal_object_id: str | None = None
        identity_profile = mc.identity_profile
        if identity_profile and isinstance(identity_profile, dict):
            kubelet = identity_profile.get("kubeletidentity")
            if kubelet is not None:
                principal_object_id = kubelet.object_id
                if not principal_object_id:
                    principal_object_id = getattr(
                        kubelet, "principal_id", None
                    )
        if not principal_object_id:
            principal_object_id = (
                mc.identity.principal_id if mc.identity else None
            )
        if not principal_object_id:
            raise RuntimeError(
                "Could not determine AKS principal to assign AcrPull."
            )

        # Assign AcrPull role (built-in role id)
        auth_client = AuthorizationManagementClient(
            credential=self._get_credential(), subscription_id=subscription_id
        )
        role_def_id = (
            f"/subscriptions/{subscription_id}/providers/"
            f"Microsoft.Authorization/roleDefinitions/"
            f"7f951dda-4ed3-4680-a7ca-43fe172d538d"
        )  # AcrPull
        assignment_name = str(uuid.uuid4())
        from azure.mgmt.authorization.models import RoleAssignmentProperties

        params = RoleAssignmentCreateParameters(
            properties=RoleAssignmentProperties(
                role_definition_id=role_def_id,
                principal_id=principal_object_id,
            )
        )
        try:
            auth_client.role_assignments.create(scope, assignment_name, params)
        except Exception as e:
            # Ignore conflicts if role already assigned
            msg = str(e).lower()
            if "conflict" in msg or "already exists" in msg:
                return
            raise

    def get_resource(
        self,
        resource_group: str,
        name: str,
        subscription_id: str | None = None,
    ) -> Response[AzureKubernetesServiceResource]:
        if not name:
            raise BadRequestError("Cluster name is required.")
        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.containerservice import ContainerServiceClient

        client = ContainerServiceClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        try:
            result = client.managed_clusters.get(resource_group, name)
            resource = AzureKubernetesServiceResource(
                id=result.id,
                name=result.name,
                location=result.location,
                kubernetes_version=result.kubernetes_version,
                fqdn=result.fqdn or result.private_fqdn,
                dns_prefix=result.dns_prefix,
            )
        except ResourceNotFoundError:
            raise NotFoundError("Kubernetes cluster not found.")
        return Response(result=resource)

    def delete_resource(
        self,
        resource_group: str,
        name: str,
        subscription_id: str | None = None,
        delete_empty_resource_group: bool = False,
    ) -> Response[None]:
        if not name:
            raise BadRequestError("Cluster name is required.")

        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.containerservice import ContainerServiceClient

        client = ContainerServiceClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        try:
            poller = client.managed_clusters.begin_delete(resource_group, name)
            poller.wait()
        except ResourceNotFoundError:
            raise NotFoundError("Kubernetes cluster not found.")

        if delete_empty_resource_group:
            self._delete_resource_group_if_empty(
                resource_group=resource_group,
                subscription_id=subscription_id,
            )
        return Response(result=None)

    def _get_provider_kubeconfig(self) -> str | dict[str, Any] | None:
        import base64

        import yaml
        from azure.mgmt.containerservice import ContainerServiceClient

        if self.resource_group is None:
            raise BadRequestError("Resource group is required.")
        if self.name is None:
            raise BadRequestError("Cluster name is required.")

        client = ContainerServiceClient(
            credential=self._get_credential(),
            subscription_id=self._get_default_subscription_id(),
        )
        k = client.managed_clusters.list_cluster_user_credentials(
            resource_group_name=self.resource_group, resource_name=self.name
        )
        if not k.kubeconfigs or len(k.kubeconfigs) == 0:
            raise NotFoundError("No kubeconfig found for the cluster.")
        kubeconfig_b64 = k.kubeconfigs[0].value  # bytes (base64)
        if kubeconfig_b64 is None:
            raise NotFoundError("No kubeconfig value found for the cluster.")
        kubeconfig_yaml = base64.b64decode(kubeconfig_b64).decode("utf-8")
        kubeconfig_obj = yaml.safe_load(kubeconfig_yaml)

        return kubeconfig_obj



"""
Local Docker container deployment provider.
"""

from __future__ import annotations

from typing import Literal

__all__ = ["DockerLocal"]

import time
from typing import Any

import docker
from docker.errors import NotFound
from docker.models.containers import Container as DockerContainer
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Context, OperationParser, Response, RunContext
from x8.core.constants import ROOT_PACKAGE_NAME
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import Expression

from .._feature import ContainerDeploymentFeature
from .._helper import map_images, requires_container_registry
from .._models import (
    Container,
    EnvVar,
    Ingress,
    Port,
    ResourceLimits,
    ResourceRequirements,
    Revision,
    Scale,
    ServiceDefinition,
    ServiceItem,
    ServiceOverlay,
    TrafficAllocation,
    VolumeMount,
)
from ._base import BaseContainerDeploymentProvider

LABEL_PREFIX = ROOT_PACKAGE_NAME


class DockerLocal(BaseContainerDeploymentProvider):
    service_name: str | None

    _client: docker.DockerClient
    _init: bool
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        service_name: str | None = None,
        service: ServiceDefinition | None = None,
        service_override: ServiceOverlay | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize Local Docker provider.

        Args:
            service_name:
                Cloud run service name.
            service:
                Service definition to deploy.
            service_override:
                Service override definition to apply.
            containerizer:
                Containerizer to use for building the container.
            container_registry:
                Container registry to use for pushing the container.
        """
        self.service_name = service_name
        self.service = service
        self.overlay = service_override
        self.containerizer = containerizer
        self.container_registry = container_registry

        self._init = False
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()

        super().__init__(**kwargs)

    def __supports__(self, feature):
        return feature not in [
            ContainerDeploymentFeature.MULTIPLE_REVISIONS,
            ContainerDeploymentFeature.REVISION_DELETE,
        ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        self._client = docker.from_env()
        self._init = True

    def _ensure_container_registry(
        self, service: ServiceDefinition
    ) -> ContainerRegistry | None:
        container_registry = self._get_container_registry()
        if container_registry:
            return container_registry
        if not requires_container_registry(service):
            return None

        from x8.compute.container_registry.providers.docker_local import (  # noqa
            DockerLocal,
        )

        return ContainerRegistry(__provider__=DockerLocal())

    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        self.__setup__()

        where_exists = OperationParser.parse_where_exists(where)
        service_def = self._normalize_service_definition(service)
        service_name = self._get_service_name(service_def)
        existing_containers = self._get_service_containers(service_name)

        if existing_containers and where_exists is False:
            raise PreconditionFailedError(
                f"Service {service_def.name} already exists."
            )

        if not existing_containers and where_exists is True:
            raise PreconditionFailedError(
                f"Service {service_def.name} not found."
            )

        images = map_images(
            containers=service_def.containers,
            images=service_def.images,
            containerizer=self._get_containerizer(),
            container_registry=self._ensure_container_registry(service_def),
            run_context=run_context,
        )

        self._ensure_network(service_name)

        if existing_containers:
            self._stop_and_remove_containers(existing_containers)

        revision = "revision:" + str(int(time.time()))
        container_specs = (
            self._op_converter.convert_service_to_container_specs(
                service_name,
                service_def,
                images,
                revision,
            )
        )
        self._create_containers_from_specs(container_specs)
        result = self._result_converter.convert_service_item(
            service_name, service_def
        )

        action = "updated" if existing_containers else "created"
        print(f"Service {action} successfully: {service_def.name}")

        return Response(result=result)

    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        return self.create_service(
            service=service,
            where=where,
            run_context=run_context,
            **kwargs,
        )

    def get_service(self, name: str) -> Response[ServiceItem]:
        self.__setup__()

        containers = self._get_service_containers(name)
        if not containers:
            raise NotFoundError(f"Service {name} not found.")

        service_def = self._result_converter.convert_service(containers)
        result = self._result_converter.convert_service_item(
            name,
            service_def,
        )

        return Response(result=result)

    def delete_service(self, name: str) -> Response[None]:
        self.__setup__()

        containers = self._get_service_containers(name)
        if not containers:
            raise NotFoundError(f"Service {name} not found.")

        self._stop_and_remove_containers(containers)
        self._delete_network(name)
        return Response(result=None)

    def list_services(self) -> Response[list[ServiceItem]]:
        self.__setup__()

        containers = self._client.containers.list(
            all=True, filters={"label": f"{LABEL_PREFIX}.managed=true"}
        )

        services: dict = {}
        for container in containers:
            service_name = container.labels.get(f"{LABEL_PREFIX}.service")
            if service_name:
                if service_name not in services:
                    services[service_name] = []
                services[service_name].append(container)

        result = []
        for service_name, service_containers in services.items():
            service_def = self._result_converter.convert_service(
                service_containers
            )
            service_item = self._result_converter.convert_service_item(
                service_name, service_def
            )
            result.append(service_item)

        return Response(result=result)

    def list_revisions(self, name: str) -> Response[list[Revision]]:
        self.__setup__()

        containers = self._get_service_containers(name)
        if not containers:
            raise NotFoundError(f"Service {name} not found.")

        revisions = []
        revision = self._result_converter.convert_revision(containers)
        revisions.append(revision)
        return Response(result=revisions)

    def get_revision(self, name: str, revision: str) -> Response[Revision]:
        self.__setup__()

        try:
            containers = self._get_revision_containers(name, revision)
            result = self._result_converter.convert_revision(containers)
            return Response(result=result)
        except NotFound:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )

    def delete_revision(self, name: str, revision: str) -> Response[None]:
        self.__setup__()
        containers = self._get_revision_containers(name, revision)
        if not containers:
            raise NotFoundError(
                f"Revision {revision} of service {name} not found."
            )
        raise PreconditionFailedError(
            (
                f"Cannot delete active revision {revision} "
                f"of service {name}."
            )
        )

    def update_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
    ) -> Response[ServiceItem]:
        return self.get_service(name)

    def close(self) -> None:
        if hasattr(self, "_client"):
            self._client.close()
        self._init = False

    def _create_containers_from_specs(
        self, container_specs: list[dict[str, Any]]
    ) -> list[DockerContainer]:
        containers = []

        # Separate init and regular containers
        init_specs = [spec for spec in container_specs if spec.get("is_init")]
        main_specs = [
            spec for spec in container_specs if not spec.get("is_init")
        ]

        # Run init containers first and wait for completion
        for spec in init_specs:
            container = self._create_single_container(spec)
            container.wait()  # Wait for init container to complete
            containers.append(container)

        # find the main spec, run it first
        main_idx = next(
            (
                i
                for i, s in enumerate(main_specs)
                if s.get("labels", {}).get(f"{LABEL_PREFIX}.type") == "main"
            ),
            0,
        )
        main_spec = main_specs.pop(main_idx)
        main_container = self._create_single_container(main_spec)
        containers.append(main_container)
        time.sleep(5)  # Give time for the main container to start

        # Run regular containers
        for spec in main_specs:
            if spec.get("labels", {}).get(f"{LABEL_PREFIX}.type") == "side":
                spec = spec.copy()
                spec.pop("network", None)  # can't combine with network_mode
                spec.pop("ports", None)  # can't publish when sharing netns
                spec["network_mode"] = f"container:{main_container.id}"
            container = self._create_single_container(spec)
            containers.append(container)
            time.sleep(5)

        return containers

    def _create_single_container(
        self, spec: dict[str, Any]
    ) -> DockerContainer:
        """Create a single Docker container from specification."""
        # Remove custom fields that aren't Docker API parameters
        docker_spec = spec.copy()
        docker_spec.pop("is_init", None)

        return self._client.containers.run(**docker_spec)

    def _get_service_containers(
        self, service_name: str
    ) -> list[DockerContainer]:
        try:
            containers = self._client.containers.list(
                all=True,
                filters={
                    "label": [
                        f"{LABEL_PREFIX}.managed=true",
                        f"{LABEL_PREFIX}.service={service_name}",
                    ]
                },
            )
            return containers
        except Exception:
            return []

    def _get_revision_containers(
        self, service_name: str, revision: str
    ) -> list[DockerContainer]:
        try:
            containers = self._client.containers.list(
                all=True,
                filters={
                    "label": [
                        f"{LABEL_PREFIX}.managed=true",
                        f"{LABEL_PREFIX}.service={service_name}",
                        f"{LABEL_PREFIX}.revision={revision}",
                    ]
                },
            )
            return containers
        except Exception:
            return []

    def _stop_and_remove_containers(
        self, containers: list[DockerContainer]
    ) -> None:
        """Stop and remove containers."""
        for container in containers:
            try:
                container.stop(timeout=10)
                container.remove()
            except Exception as e:
                print(f"Error removing container {container.name}: {e}")

    def _ensure_network(self, service_name) -> None:
        try:
            self._client.networks.get(service_name)
        except NotFound:
            self._client.networks.create(
                service_name,
                driver="bridge",
                labels={f"{LABEL_PREFIX}.managed": "true"},
            )

    def _delete_network(self, service_name: str) -> None:
        try:
            network = self._client.networks.get(service_name)
            network.remove()
        except NotFound:
            pass

    def _get_service_name(self, service: ServiceDefinition) -> str:
        if self.service_name:
            return self.service_name
        if service.name:
            return service.name
        raise BadRequestError(
            "Service name must be provided or defined in the service."
        )


class OperationConverter:
    def convert_service_to_container_specs(
        self,
        service_name: str,
        service: ServiceDefinition,
        images: list[str],
        revision: str,
    ) -> list[dict[str, Any]]:
        container_specs = []

        for container_def, image in zip(service.containers, images):
            if container_def.type == "main" and container_def.ports:
                inferred_type = "main"
            elif container_def.type == "init":
                inferred_type = "init"
            else:
                inferred_type = "side"
            spec = self._convert_container_to_spec(
                service_name,
                service,
                container_def,
                image,
                inferred_type,
                revision,
            )
            container_specs.append(spec)

        return container_specs

    def _convert_container_to_spec(
        self,
        service_name: str,
        service: ServiceDefinition,
        container_def: Container,
        image: str,
        inferred_type: str,
        revision: str,
    ) -> dict[str, Any]:
        """Convert a single container definition to Docker spec."""
        is_init = container_def.type == "init"

        # Prepare environment variables
        environment = {}
        for env_var in container_def.env:
            environment[env_var.name] = str(env_var.value)
        explicit_env_names = ",".join(sorted(environment.keys()))

        # Prepare port bindings
        ports: dict = {}
        if container_def.ports and not is_init:
            for port in container_def.ports:
                container_port = f"{port.container_port}/tcp"
                ports[container_port] = port.host_port or port.container_port

        # Prepare volumes
        volumes = {}
        if container_def.volume_mounts:
            for mount in container_def.volume_mounts:
                volumes[mount.name] = {"bind": mount.mount_path, "mode": "rw"}

        # Prepare labels
        labels = {
            f"{LABEL_PREFIX}.managed": "true",
            f"{LABEL_PREFIX}.service": service.name,
            f"{LABEL_PREFIX}.container": container_def.name,
            f"{LABEL_PREFIX}.type": inferred_type,
            f"{LABEL_PREFIX}.created": str(int(time.time())),
            f"{LABEL_PREFIX}.env_names": explicit_env_names,
            f"{LABEL_PREFIX}.revision": revision,
        }

        # Prepare resource constraints
        mem_limit = None
        cpu_period = None
        cpu_quota = None

        if container_def.resources and container_def.resources.limits:
            if container_def.resources.limits.memory:
                mem_limit = container_def.resources.limits.memory
            if container_def.resources.limits.cpu:
                cpu_period, cpu_quota = self._convert_cpu_limit(
                    container_def.resources.limits.cpu
                )

        args = {
            "image": image,
            "entrypoint": container_def.command,
            "command": container_def.args,
            "environment": environment,
            "volumes": volumes,
            "labels": labels,
            "network": service_name,
            "name": f"{service.name}-{container_def.name}-{int(time.time())}",
            "detach": True,
            "mem_limit": mem_limit,
            "cpu_period": cpu_period,
            "cpu_quota": cpu_quota,
            "working_dir": container_def.working_dir,
            "restart_policy": (
                {"Name": "unless-stopped"} if not is_init else None
            ),
            "is_init": is_init,
        }
        if ports:
            args["ports"] = ports
        return args

    def _convert_cpu_limit(
        self, cpu_limit: float
    ) -> tuple[int | None, int | None]:
        """Convert CPU limit to Docker CPU period and quota."""
        cores = float(cpu_limit)
        cpu_period = 100000
        cpu_quota = int(cores * cpu_period)
        return cpu_period, cpu_quota


class ResultConverter:
    def convert_service_item(
        self,
        service_name: str,
        service_def: ServiceDefinition,
    ) -> ServiceItem:
        # Generate a simple URI for the service
        uri = "http://localhost"
        return ServiceItem(
            name=service_name,
            uri=uri,
            service=service_def,
        )

    def convert_service(
        self,
        containers: list[DockerContainer],
    ) -> ServiceDefinition:
        # Get service name from first container
        main = next(
            (
                c
                for c in containers
                if c.labels.get(f"{LABEL_PREFIX}.type") == "main"
            ),
            containers[0],
        )
        service_name = main.labels.get(f"{LABEL_PREFIX}.service")
        revision = main.labels.get(f"{LABEL_PREFIX}.revision", "unknown")
        if not service_name:
            raise ValueError("Service name not found in container labels")

        # Convert containers
        container_defs = []
        for docker_container in containers:
            container_def = self._convert_container(docker_container)
            container_defs.append(container_def)

        ingress = Ingress(external=True)
        traffic = [
            TrafficAllocation(
                revision=revision,
                percent=100,
                latest_revision=True,
            )
        ]

        # Basic scaling info
        scale = Scale(
            mode="manual",
            replicas=1,
        )

        return ServiceDefinition(
            name=service_name,
            containers=container_defs,
            ingress=ingress,
            traffic=traffic,
            scale=scale,
            latest_ready_revision=revision,
            latest_created_revision=revision,
        )

    def _convert_container(
        self, docker_container: DockerContainer
    ) -> Container:
        """Convert Docker container to Container definition."""

        # Reload container to get fresh state
        docker_container.reload()

        container_name = docker_container.labels.get(
            f"{LABEL_PREFIX}.container", "unknown"
        )
        container_type = docker_container.labels.get(
            f"{LABEL_PREFIX}.type", "regular"
        )

        # Extract environment variables
        env_vars = []
        labels = docker_container.labels or {}
        explicit = labels.get(f"{LABEL_PREFIX}.env_names", "")
        explicit_set = {n for n in explicit.split(",") if n}
        cfg_env = (docker_container.attrs.get("Config", {}) or {}).get(
            "Env"
        ) or []
        for env_str in cfg_env:
            if "=" not in env_str:
                continue
            name, value = env_str.split("=", 1)
            if name in explicit_set:
                env_vars.append(EnvVar(name=name, value=value))

        # Extract ports
        ports = []
        if docker_container.attrs.get("Config", {}).get("ExposedPorts"):
            for port_str in docker_container.attrs["Config"][
                "ExposedPorts"
            ].keys():
                if "/" in port_str:
                    port_num = int(port_str.split("/")[0])
                    ports.append(
                        Port(container_port=port_num, name=f"port-{port_num}")
                    )

        # Extract resource requirements
        resources = None
        config = docker_container.attrs.get("HostConfig", {})
        if config.get("Memory") or config.get("CpuQuota"):
            limits = ResourceLimits()
            if config.get("Memory"):
                limits.memory = config["Memory"]
            if config.get("CpuQuota") and config.get("CpuPeriod"):
                cpu_cores = config["CpuQuota"] / config["CpuPeriod"]
                limits.cpu = cpu_cores
            resources = ResourceRequirements(limits=limits)

        # Extract volumes
        volume_mounts = []
        if docker_container.attrs.get("Mounts"):
            for mount in docker_container.attrs["Mounts"]:
                if mount.get("Type") == "volume":
                    volume_mounts.append(
                        VolumeMount(
                            name=mount.get(
                                "Name", mount.get("Source", "unknown")
                            ),
                            mount_path=mount["Destination"],
                        )
                    )
        type: Literal["main", "init"] = (
            "init" if container_type == "init" else "main"
        )
        return Container(
            name=container_name,
            image=(
                docker_container.image.tags[0]
                if docker_container.image.tags
                else "unknown"
            ),
            command=docker_container.attrs.get("Config", {}).get(
                "Entrypoint", []
            ),
            args=docker_container.attrs.get("Config", {}).get("Cmd", []),
            working_dir=docker_container.attrs.get("Config", {}).get(
                "WorkingDir"
            ),
            env=env_vars,
            ports=ports,
            resources=resources,
            volume_mounts=volume_mounts,
            type=type,
        )

    def convert_revision(self, containers: list[DockerContainer]) -> Revision:
        """Convert Docker container to Revision."""
        container_defs = []
        for docker_container in containers:
            container_def = self._convert_container(docker_container)
            container_defs.append(container_def)

        main = next(
            (
                c
                for c in containers
                if c.labels.get(f"{LABEL_PREFIX}.type") == "main"
            ),
            containers[0],
        )

        # Get creation time
        created_time = None
        if main.attrs.get("Created"):
            from datetime import datetime

            created_str = main.attrs["Created"]
            # Parse ISO format timestamp
            created_dt = datetime.fromisoformat(
                created_str.replace("Z", "+00:00")
            )
            created_time = created_dt.timestamp()

        # Check if container is active (running)
        active = main.status == "running"
        revision = main.labels.get(f"{LABEL_PREFIX}.revision", "unknown")

        return Revision(
            name=revision,
            traffic=100 if active else 0,
            created_time=created_time,
            status="active" if active else "inactive",
            active=active,
            containers=container_defs,
            volumes=[],
        )

    def convert_uri(self, service: ServiceDefinition) -> str:
        host_port = None
        for container in service.containers:
            if container.type == "main":
                if container.ports:
                    host_port = (
                        container.ports[0].host_port
                        or container.ports[0].container_port
                    )
                    break
        uri = (
            f"http://localhost:{host_port}"
            if host_port
            else "http://localhost"
        )
        return uri



from typing import Any, AsyncIterator, Iterator, List, Literal, overload

from x8.core import Component, Response, operation

from ._models import (
    ImageData,
    KeyFrame,
    Reference,
    VideoGenerationResult,
    VideoSize,
)


class VideoGeneration(Component):
    @overload
    def generate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: Literal[False] | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response[VideoGenerationResult]:
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @overload
    def generate(
        self,
        prompt: str | None = None,
        *,
        stream: Literal[True],
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Iterator[Response[VideoGenerationResult]]:
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "",
            "method": "POST",
            "status": 201,
        }
    )
    def generate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[VideoGenerationResult]
        | Iterator[Response[VideoGenerationResult]]
    ):
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}",
            "method": "GET",
            "status": 200,
        }
    )
    def get(self, id: str, **kwargs: Any) -> Response[VideoGenerationResult]:
        """
        Get the status and details of a video generation task.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "",
            "method": "GET",
            "status": 200,
        }
    )
    def list(self, **kwargs: Any) -> Response[List[VideoGenerationResult]]:
        """
        List all video generation tasks.

        Args:
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}",
            "method": "DELETE",
            "status": 204,
        }
    )
    def delete(self, id: str, **kwargs: Any) -> Response[None]:
        """
        Delete a video generation task.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}/content",
            "method": "GET",
            "status": 200,
            "response": {
                "media_type": "video/mp4",
            },
        }
    )
    def download(self, id: str, **kwargs: Any) -> Iterator[bytes]:
        """
        Download the generated video.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @overload
    async def agenerate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: Literal[False] | None = None,
        nconfig: dict[str, Any] | None = None,
    ) -> Response[VideoGenerationResult]:
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @overload
    async def agenerate(
        self,
        prompt: str | None = None,
        *,
        stream: Literal[True],
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        nconfig: dict[str, Any] | None = None,
    ) -> AsyncIterator[Response[VideoGenerationResult]]:
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "",
            "method": "POST",
            "status": 201,
            "response": {"type": "stream"},
        }
    )
    async def agenerate(
        self,
        prompt: str | None = None,
        *,
        image: ImageData | None = None,
        references: List[Reference] | None = None,
        key_frames: List[KeyFrame] | None = None,
        negative_prompt: str | None = None,
        audio: bool | None = True,
        duration: float | None = None,
        size: VideoSize | None = None,
        samples: int | None = None,
        seed: int | None = None,
        quality: str | None = None,
        poll: bool = False,
        poll_interval: float = 5.0,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[VideoGenerationResult]
        | AsyncIterator[Response[VideoGenerationResult]]
    ):
        """
        Generate a video from input.

        Args:
            prompt:
                The text prompt to generate the video.
            image:
                An image to guide the video generation.
            references:
                List of reference images to guide the video generation.
            key_frames:
                List of key frames to include in the video.
            negative_prompt:
                The negative text prompt to avoid in the video.
            audio:
                Whether to include audio in the generated video.
            duration:
                The duration of the video in seconds.
            size:
                The size of the video.
            samples:
                The number of video samples to generate.
            seed:
                The random seed for reproducibility.
            quality:
                The quality setting for the video generation.
            poll:
                Whether to poll for completion.
            poll_interval:
                The interval (in seconds) between polling attempts.
            stream:
                Whether to stream the response.
            nconfig:
                Additional native configuration parameters.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}",
            "method": "GET",
            "status": 200,
        }
    )
    async def aget(
        self, id: str, **kwargs: Any
    ) -> Response[VideoGenerationResult]:
        """
        Get the status and details of a video generation task.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "",
            "method": "GET",
            "status": 200,
        }
    )
    async def alist(
        self, **kwargs: Any
    ) -> Response[List[VideoGenerationResult]]:
        """
        List all video generation tasks.

        Args:
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}",
            "method": "DELETE",
            "status": 204,
        }
    )
    async def adelete(self, id: str, **kwargs: Any) -> Response[None]:
        """
        Delete a video generation task.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError

    @operation(
        api={
            "path": "/{id}/content",
            "method": "GET",
            "status": 200,
            "response": {
                "media_type": "video/mp4",
            },
        }
    )
    async def adownload(self, id: str, **kwargs: Any) -> AsyncIterator[bytes]:
        """
        Download the generated video.

        Args:
            id:
                The ID of the video generation task.
            **kwargs: Additional keyword arguments.
        """
        raise NotImplementedError



from ._providers import SearchStoreProvider

complex_condition_1 = """
    length(arrstr[0]) > 7
    and not contains(const, '$')
    and 8 = array_length(obj.narr)
    and array_contains(arrint, 909)
    and array_contains_any(arrstr, ['xyz', 'hundred nine', 'abc'])
    and is_defined(str)
    and is_not_defined(opt)
    and is_type(float, 'number')
    and is_type(empty, 'null')
    and not is_type(bool, 'array')
    and obj.nstr = "9"
    and 9.1 != float
    and int >= 8 and (bool = true or obj.nobj.nnfloat <= -900.1)
    and not (pk = "pk00" or length(obj.nobj.nnstr) != 2)
    and obj.nint between -10 and 10
    and str in ('one', 'two', 'eight', 'nine')
    and obj.nint not in (-1, -2, -8)
    """
bad_complex_condition_1 = f"{complex_condition_1} and obj.nint > 0"

complex_condition_1_with_params = """
    length(arrstr[0]) > @p1
    and not contains(const, @p3)
    and @p5 = array_length(obj.narr)
    and array_contains(arrint, @p6)
    and array_contains_any(arrstr, @p7)
    and is_defined(str)
    and is_not_defined(opt)
    and is_type(float, @p8)
    and is_type(empty, @p10)
    and not is_type(bool, @p11)
    and obj.nstr = @p12
    and @p14 != float
    and int >= @p15 and (bool = @p16 or obj.nobj.nnfloat <= @p17)
    and not (pk = @p18 or length(obj.nobj.nnstr) != @p19)
    and obj.nint between @p20 and @p21
    and str in (@p22, @p23, @p24, @p25)
    and obj.nint not in (@p26, @p27, @p28)
    """

complex_condition_1_params = {
    "p1": 7,
    "p2": "und",
    "p3": "$",
    "p4": "nine",
    "p5": 8,
    "p6": 909,
    "p7": ["xyz", "hundred nine", "abc"],
    "p8": "number",
    "p9": "object",
    "p10": "null",
    "p11": "array",
    "p12": "9",
    "p13": 9000000000,
    "p14": 9.1,
    "p15": 8,
    "p16": True,
    "p17": -900.1,
    "p18": "pk00",
    "p19": 2,
    "p20": -10,
    "p21": 10,
    "p22": "one",
    "p23": "two",
    "p24": "eight",
    "p25": "nine",
    "p26": -1,
    "p27": -2,
    "p28": -8,
}

complex_condition_2 = """
    length(arrstr[0]) > 7
    and not contains(const, '$')
    and 7 = array_length(obj.narr)
    and array_contains(arrint, 808)
    and array_contains_any(arrstr, ['xyz', 'hundred eight', 'abc'])
    and is_defined(str)
    and is_not_defined(opt)
    and is_type(float, 'number')
    and is_type(empty, 'null')
    and not is_type(bool, 'array')
    and obj.nstr = "8"
    and 8.1 != float
    and int >= 7 and (bool = false or obj.nobj.nnfloat <= -800.1)
    and not (pk = "pk00" or length(obj.nobj.nnstr) != 4)
    and obj.nint between -10 and 10
    and str in ('one', 'two', 'eight', 'nine')
    and obj.nint not in (-1, -2, -9)
    """

queries = [
    {
        "args": {
            "where": "pk = 'pk00'",
            "order_by": "id",
        },
        "statement": "WHERE pk = 'pk00' ORDER BY id",
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "select": "*",
            "where": "pk = 'pk00'",
            "order_by": "$id",
        },
        "statement": "SELECT * WHERE pk = 'pk00' ORDER BY $id",
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "select": "bool, obj, pk, $id",
            "where": "pk = 'pk00'",
            "order_by": "id",
        },
        "statement": """SELECT bool, obj, pk, $id
                    WHERE pk = 'pk00' ORDER BY $id""",
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "select": "obj.nstr, obj.nint, pk, $id",
            "where": "pk = 'pk00'",
            "order_by": "$id",
        },
        "statement": "SELECT @p1 WHERE @p2 ORDER BY @p3",
        "params": {
            "p1": "obj.nstr, obj.nint, pk, $id",
            "p2": "pk = @p4",
            "p3": "$id",
            "p4": "pk00",
        },
        "result_index": [0, 1, 2, 3, 4],
        "count": 5,
    },
    {
        "args": {
            "where": "pk = 'pk01'",
            "order_by": "id DESC",
        },
        "statement": "WHERE pk = 'pk01' ORDER BY id DESC",
        "result_index": [9, 8, 7, 6, 5],
        "count": 5,
    },
    {
        "args": {
            "where": "pk = 'pk01'",
            "order_by": "id DESC",
            "limit": 3,
        },
        "statement": "WHERE pk = 'pk01' ORDER BY id DESC LIMIT 3",
        "result_index": [9, 8, 7],
        "count": 5,
    },
    {
        "args": {
            "where": "pk = 'pk01'",
            "order_by": "id DESC",
            "limit": 10,
            "offset": 3,
        },
        "statement": "WHERE pk = 'pk01' ORDER BY id DESC LIMIT 10 OFFSET 3",
        "result_index": [6, 5],
        "count": 5,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and int <= 2
                """,
            "order_by": "int",
        },
        "statement": "WHERE pk = 'pk00' and int <=2 ORDER BY int",
        "result_index": [0, 1, 2],
        "count": 3,
    },
    {
        "args": {
            "where": """
                pk = 'pk01' and str > 'eight'
                """,
            "order_by": "str",
        },
        "statement": "WHERE pk = 'pk01' and str > 'eight' ORDER BY str",
        "result_index": [5, 9, 7, 6],
        "count": 4,
        "except_providers": [],
    },
    {
        "args": {
            "where": """
                 pk = 'pk01' and int != 7
                 """
        },
        "statement": "WHERE pk = 'pk01' and int !=7",
        "result_index": [9, 8, 6, 5],
        "ordered": False,
        "count": 4,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and int between 1 and 3
                """,
            "order_by": "int DESC",
        },
        "statement": """WHERE pk = 'pk00' and int between 1 and 3
                        ORDER BY int DESC""",
        "result_index": [3, 2, 1],
        "count": 3,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and opt in ('abcd', 'cdef', 'xxx')
                """
        },
        "statement": """WHERE pk = 'pk00'
                    and opt in ('abcd', 'cdef', 'xxx')
                    """,
        "result_index": [2, 4],
        "ordered": False,
        "count": 2,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and int not in (1, 3, 5)
                """
        },
        "statement": "WHERE pk = 'pk00' and int not in (1, 3, 5)",
        "result_index": [0, 2, 4],
        "ordered": False,
        "count": 3,
    },
    {
        "args": {
            "where": """
                pk = 'pk01'
                and str = 'xsix'
                and (int = 5 or bool = true)
                """,
            "order_by": "str",
        },
        "statement": """WHERE
                pk = 'pk01'
                and str = 'xsix'
                and (int = 5 or bool = true)
                ORDER BY str
                """,
        "result_index": [6],
        "count": 1,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and obj.nint >= -1
                """
        },
        "statement": """WHERE
                pk = 'pk00' and obj.nint >= -1
                """,
        "result_index": [0, 1],
        "ordered": False,
        "count": 2,
    },
    {
        "args": {
            "where": """
                pk = 'pk01' and starts_with(str, 'xs')
                """,
            "order_by": "str",
        },
        "statement": """WHERE
                pk = 'pk01' and starts_with(str, 'xs')
                ORDER BY str
                """,
        "result_index": [7, 6],
        "count": 2,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and array_contains(arrstr, 'zero')
                """,
            "order_by": "str",
        },
        "statement": """WHERE
                pk = 'pk00' and array_contains(arrstr, 'zero')
                ORDER BY str
                """,
        "result_index": [4, 3, 2, 0],
        "count": 4,
    },
    {
        "args": {
            "where": """
                pk = 'pk01'
                and array_contains(arrobj,
                {"ostr": "nine million", "oint": 9000000})
                """,
            "order_by": "str",
        },
        "statement": """WHERE
                pk = 'pk01'
                and array_contains(arrobj,
                {"ostr": "nine million", "oint": 9000000})
                ORDER BY str
                """,
        "result_index": [9],
        "count": 1,
        "except_providers": [SearchStoreProvider.ELASTICSEARCH],
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and array_contains_any(obj.narr, [400, 200, 1000])
                """,
            "order_by": "int DESC",
        },
        "statement": "WHERE @p1 ORDER BY @p2",
        "params": {
            "p1": """pk = 'pk00'
                    and array_contains_any(obj.narr, [400, 200, 1000])""",
            "p2": "int DESC",
        },
        "result_index": [4, 2],
        "count": 2,
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and contains(opt, 'cde')
                """,
            "order_by": "int",
        },
        "statement": "WHERE @p1 ORDER BY @p2",
        "params": {"p1": "pk = 'pk00' and contains(opt, 'cde')", "p2": "int"},
        "result_index": [3, 4],
        "count": 2,
        "except_providers": [],
    },
    {
        "args": {
            "where": """
                pk = 'pk00' and arrstr[3] = 'zero'
                and obj.narr[0] <= 200
                and arrobj[1].oint >= 2000000000
                """
        },
        "statement": """WHERE
                pk = 'pk00' and arrstr[3] = 'zero'
                and obj.narr[0] <= 200
                and arrobj[1].oint >= 2000000000
                """,
        "result_index": [2],
        "ordered": False,
        "count": 1,
        "except_providers": [SearchStoreProvider.ELASTICSEARCH],
    },
    {
        "args": {"where": complex_condition_1},
        "statement": f"WHERE {complex_condition_1}",
        "result_index": [9],
        "count": 1,
        "except_providers": [],
    },
    {
        "args": {"where": complex_condition_2},
        "statement": f"WHERE {complex_condition_2}",
        "result_index": [8],
        "count": 1,
        "except_providers": [],
    },
]

text_search_queries = [
    {
        "args": {
            "search": 'text_search("yellow")',
            "where": "pk = 'pk00'",
            "order_by": "id",
        },
        "statement": """SEARCH text_search("yellow")
                        WHERE pk = 'pk00' ORDER BY id""",
        "result_index": [0, 2],
        "count": 2,
    },
    {
        "args": {
            "search": 'text_search("yellow")',
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search("yellow")
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2, 0],
        "count": 2,
    },
    {
        "args": {
            "search": 'text_search("yellow")',
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search("yellow")
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2, 0],
        "count": 2,
    },
    {
        "args": {
            "search": 'text_search("yellow code")',
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search("yellow code")
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2, 0],
        "count": 2,
    },
    {
        "args": {
            "search": 'text_search(query="yellow code", match_mode="and")',
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow code", match_mode="and")
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2],
        "count": 1,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood", match_mode="and")',
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood", match_mode="and")
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [],
        "count": 0,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood", match_mode="and", query_type="cross_fields")',  # noqa
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood", match_mode="and", query_type="cross_fields") # noqa
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2],
        "count": 1,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood", fields=["text1"], match_mode="and", query_type="cross_fields")',  # noqa
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood", fields=["text1"], match_mode="and", query_type="cross_fields") # noqa
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [],
        "count": 0,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood", fields=["text1", "text2"], match_mode="and", query_type="cross_fields")',  # noqa
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood", fields=["text1", "text2"], match_mode="and", query_type="cross_fields") # noqa
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [2],
        "count": 1,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood green blue", fields=["text1", "text2"], match_mode="and", query_type="cross_fields", minimum_should_match="75%")',  # noqa
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood green blue", fields=["text1", "text2"], match_mode="and", query_type="cross_fields", minimum_should_match="75%") # noqa
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [],
        "count": 0,
    },
    {
        "args": {
            "search": 'text_search(query="yellow wood green blue", fields=["text1", "text2"], match_mode="and", query_type="cross_fields", minimum_should_match="50%")',  # noqa
            "where": "pk = 'pk00'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="yellow wood green blue", fields=["text1", "text2"], match_mode="and", query_type="cross_fields", minimum_should_match="50%") # noqa
                        WHERE pk = 'pk00' ORDER BY $score DESC""",
        "result_index": [],
        "count": 0,
    },
    {
        "args": {
            "search": 'text_search(query="cats arms")',  # noqa
            "where": "pk = 'pk01'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="cats arms") # noqa
                        WHERE pk = 'pk01' ORDER BY $score DESC""",
        "result_index": [8, 9],
        "count": 2,
    },
    {
        "args": {
            "search": 'text_search(query="cats arms", boost={"text1": 2, "text2": 1})',  # noqa
            "where": "pk = 'pk01'",
            "order_by": "$score DESC",
        },
        "statement": """SEARCH text_search(
                        query="cats arms", boost={"text1": 2, "text2": 1}) # noqa
                        WHERE pk = 'pk01' ORDER BY $score DESC""",
        "result_index": [9, 8],
        "count": 2,
    },
]

vector_search_queries = [
    {
        "args": {
            "search": "vector_search([1, 1, 0, 0], 'vector', 3)",
        },
        "statement": """SEARCH vector_search([1, 1, 0, 0], 'vector', 3)""",
        "result_index": [1, 0, 2],
        "count": 3,
    },
    {
        "args": {
            "search": "vector_search([4, 4, 4, 4], 'vector', 4)",
        },
        "statement": """SEARCH vector_search([4, 4, 4, 4], 'vector', 4)""",
        "result_index": [4, 5, 3, 6],
        "count": 4,
    },
    {
        "args": {
            "search": "vector_search([4, 4, 4, 4], 'vector', 4)",
            "where": "pk = 'pk00'",
        },
        "statement": """SEARCH vector_search([4, 4, 4, 4], 'vector', 4)""",
        "result_index": [4, 3, 2, 1],
        "count": 4,
    },
]

sparse_vector_search_queries = [
    {
        "args": {
            "search": 'sparse_vector_search({"0": 0.31, "3": 0.76}, "sparse_vector")',  # noqa
            "limit": 3,
        },
        "statement": """SEARCH sparse_vector_search({"0": 0.31, "3": 0.76}, "sparse_vector")""",  # noqa
        "result_index": [3, 6, 4],
        "count": 3,
    },
    {
        "args": {
            "search": 'sparse_vector_search({"0": 0.31, "3": 0.76}, "sparse_vector")',  # noqa
            "where": "pk = 'pk00'",
            "limit": 3,
        },
        "statement": """SEARCH sparse_vector_search({"0": 0.31, "3": 0.76}, "sparse_vector")""",  # noqa
        "result_index": [3, 4, 1],
        "count": 3,
    },
]

hybrid_vector_search_queries = [
    {
        "args": {
            "search": """hybrid_vector_search([1, 1, 0, 0], 'vector', 3, 10, {"0": 0.31, "3": 0.76}, 'sparse_vector')""",  # noqa
            "limit": 3,
        },
        "statement": """SEARCH hybrid_vector_search([1, 1, 0, 0], 'vector', 3, 10, {"0": 0.31, "3": 0.76}, 'sparse_vector')""",  # noqa
        "result_index": [3, 1, 6],
        "count": 3,
    },
    {
        "args": {
            "search": """hybrid_vector_search([1, 1, 0, 0], 'vector', 3, 10, {"0": 0.31, "3": 0.76}, 'sparse_vector')""",  # noqa
            "where": "pk = 'pk00'",
            "limit": 3,
        },
        "statement": """SEARCH hybrid_vector_search([1, 1, 0, 0], 'vector', 3, 10, {"0": 0.31, "3": 0.76}, 'sparse_vector')""",  # noqa
        "result_index": [3, 1, 4],
        "count": 3,
    },
]

hybrid_text_search_queries = [
    {
        "args": {
            "search": """hybrid_text_search(query="yellow wood", vector=[1, 1, 0, 0], vector_field='vector', k=3, num_candidates=10)""",  # noqa,
            "where": "pk = 'pk00'",
            "limit": 3,
        },
        "statement": """SEARCH hybrid_text_search(query="yellow wood", vector=[1, 1, 0, 0], vector_field='vector', k=3, num_candidates=10)""",  # noqa
        "result_index": [2, 0, 1],
        "count": 3,
    }
]



from x8.compute._common._models import ImageMap
from x8.ql import Expression

from ._feature import ContainerDeploymentFeature
from ._models import (
    Container,
    ContainerOverride,
    ContainerRegistryCredentials,
    EnvVar,
    ExecAction,
    GPUResource,
    GRPCAction,
    HTTPGetAction,
    HTTPHeader,
    Ingress,
    Lifecycle,
    Port,
    Probe,
    ProbeSet,
    ResourceLimits,
    ResourceRequests,
    ResourceRequirements,
    Revision,
    Scale,
    ScaleRule,
    SecurityContext,
    ServiceDefinition,
    ServiceItem,
    ServiceOverlay,
    TCPSocketAction,
    TrafficAllocation,
    Volume,
    VolumeMount,
)
from .component import ContainerDeployment

__all__ = [
    "ContainerDeployment",
    "ContainerDeploymentFeature",
    "ServiceItem",
    "ServiceDefinition",
    "ServiceOverlay",
    "Container",
    "ContainerOverride",
    "ContainerRegistryCredentials",
    "EnvVar",
    "ExecAction",
    "GPUResource",
    "GRPCAction",
    "HTTPGetAction",
    "HTTPHeader",
    "ImageMap",
    "Ingress",
    "Lifecycle",
    "Port",
    "Probe",
    "ProbeSet",
    "ResourceLimits",
    "ResourceRequests",
    "ResourceRequirements",
    "Revision",
    "Scale",
    "ScaleRule",
    "SecurityContext",
    "TCPSocketAction",
    "TrafficAllocation",
    "Volume",
    "VolumeMount",
    "Expression",
]



from typing import Any

from common.secrets import get_secrets

from x8.compute.container_deployment import ContainerDeployment

secrets = get_secrets()


class ContainerDeploymentProvider:
    AWS_APP_RUNNER = "aws_app_runner"
    AMAZON_ECS_EC2 = "amazon_ecs_ec2"
    AMAZON_ECS_FARGATE = "amazon_ecs_fargate"
    AZURE_CONTAINER_INSTANCES = "azure_container_instances"
    AZURE_CONTAINER_APPS = "azure_container_apps"
    GOOGLE_CLOUD_RUN = "google_cloud_run"
    LOCAL = "local"


provider_parameters: dict[str, dict[str, Any]] = {
    ContainerDeploymentProvider.AWS_APP_RUNNER: {},
    ContainerDeploymentProvider.AMAZON_ECS_FARGATE: {},
    ContainerDeploymentProvider.AMAZON_ECS_EC2: {
        "ec2_desired_capacity": 3,
    },
    ContainerDeploymentProvider.AZURE_CONTAINER_INSTANCES: {},
    ContainerDeploymentProvider.AZURE_CONTAINER_APPS: {},
    ContainerDeploymentProvider.GOOGLE_CLOUD_RUN: {},
    ContainerDeploymentProvider.LOCAL: {},
}


def get_component(provider_type: str):
    component = ContainerDeployment(
        __provider__=dict(
            type=provider_type,
            parameters=provider_parameters[provider_type],
        ),
    )
    return component



import multiprocessing
from typing import Any

from x8.core import (
    Component,
    Context,
    Operation,
    OperationParser,
    Provider,
    Response,
)

from .._operation import WorkerOperation


def _start(context, queue, component, max_wait_time, stop_event):
    while not stop_event.is_set():
        res = queue.__run__(
            Operation(
                name="pull",
                args=dict(config=dict(max_wait_time=max_wait_time)),
            ),
            context,
        )
        if len(res.result) > 0:
            job = res.result[0]
            component.__run__(
                job.value["operation"],
                job.value["context"],
            )
            queue.__run__(Operation(name="ack", args=dict(key=job.key)))


async def _astart(context, queue, component, max_wait_time, stop_event):
    while not stop_event.is_set():
        res = await queue.__arun__(
            Operation(
                name="pull",
                args=dict(config=dict(max_wait_time=max_wait_time)),
            ),
            context,
        )
        if len(res.result) > 0:
            job = res.result[0]
            component.__run__(
                job.value["operation"],
                job.value["context"],
            )
            await queue.__arun__(Operation(name="ack", args=dict(key=job.key)))


class Default(Provider):
    queue: Component
    component: Component
    workers: int
    max_wait_time: int
    nparams: dict[str, Any]

    _stop_event: Any
    _processes: list

    def __init__(
        self,
        queue: Component,
        component: Component,
        workers: int = 1,
        max_wait_time: int = 30,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            queue:
                Queue component from which jobs are fetched for processing.
            component:
                Component to host as API.
            workers:
                Number of worker processes.
            max_wait_time:
                Maximum wait time to receive item from the queue.
                If time expires, the worker will automatically check again
                unless the workers are stopped.
            nparams:
                Native parameters to FastAPI and uvicorn client.
        """
        self.queue = queue
        self.component = component
        self.workers = workers
        self.nparams = nparams
        self.max_wait_time = max_wait_time
        self._stop_event = multiprocessing.Event()
        self._processes = []
        print("Started worker")

    def __setup__(self, context: Context | None = None) -> None:
        self.__setup__(context=context)

    async def __asetup__(self, context: Context | None = None) -> None:
        await self.__asetup__(context=context)

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Response[None]:
        op_parser = OperationParser(operation)
        if operation is None or op_parser.op_equals(WorkerOperation.START):
            if self.workers == 1:
                _start(
                    context,
                    self.queue,
                    self.component,
                    self.max_wait_time,
                    self._stop_event,
                )
            else:
                for _ in range(self.workers):
                    print("starting process")
                    p = multiprocessing.Process(
                        target=_start,
                        kwargs=dict(
                            context=(
                                context.to_dict()
                                if context is not None
                                else None
                            ),
                            queue=self.queue,
                            component=self.component,
                            max_wait_time=self.max_wait_time,
                            stop_event=self._stop_event,
                        ),
                    )
                    p.start()
                    self._processes.append(p)
                for p in self._processes:
                    p.join()
        elif op_parser.op_equals(WorkerOperation.STOP):
            self._stop(context)
        for p in self._processes:
            p.join()
        return Response(result=None)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Response[None]:
        op_parser = OperationParser(operation)
        if operation is None or op_parser.op_equals(WorkerOperation.START):
            if self.workers == 1:
                await _astart(
                    context,
                    self.queue,
                    self.component,
                    self.max_wait_time,
                    self._stop_event,
                )
            else:
                for _ in range(self.workers):
                    p = multiprocessing.Process(
                        target=_astart,
                        kwargs=dict(
                            context=(
                                context.to_dict()
                                if context is not None
                                else None
                            ),
                            queue=self.queue,
                            component=self.component,
                            max_wait_time=self.max_wait_time,
                            stop_event=self._stop_event,
                        ),
                    )
                    p.start()
                    self._processes.append(p)
                for p in self._processes:
                    p.join()
        elif op_parser.op_equals(WorkerOperation.STOP):
            self._stop(context)
        return Response(result=None)

    def _stop(self, context):
        self._stop_event.set()



import pytest

from x8.core.exceptions import NotFoundError


def test_azure_provider():
    from x8.compute.kubernetes.providers.azure_kubernetes_service import (  # noqa
        AzureKubernetesService,
    )

    rg = "test-resource-group"
    name = "rpdtest2"
    location = "westus2"
    client = AzureKubernetesService()

    with pytest.raises(NotFoundError):
        client.get_resource(
            resource_group=rg,
            name=name,
        )

    res = client.create_resource(
        resource_group=rg,
        name=name,
        location=location,
    )
    result = res.result
    assert result is not None
    assert result.id is not None
    assert result.name == name
    assert result.location == location
    assert result.fqdn is not None

    res = client.get_resource(
        resource_group=rg,
        name=name,
    )
    result = res.result
    assert result is not None
    assert result.id is not None
    assert result.name == name
    assert result.location == location
    assert result.fqdn is not None

    client.delete_resource(
        resource_group=rg,
        name=name,
        delete_empty_resource_group=True,
    )
    with pytest.raises(NotFoundError):
        client.get_resource(
            resource_group=rg,
            name=name,
        )



"""
Image processed with Pillow.
"""

__all__ = ["Pillow"]


import base64
from io import BytesIO
from typing import Any

from PIL import Image as PILImage
from PIL.Image import Image
from x8.core import Context, Provider
from x8.core.exceptions import BadRequestError

from .._models import ImageInfo


class Pillow(Provider):
    _init: bool
    _image: Image
    _image_bytes: dict
    _image_base64: dict

    def __init__(self, **kwargs):
        """Intialize."""
        self._init = False
        self._image_bytes = dict()
        self._image_base64 = dict()

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return
        source = self.__component__.source
        url = None
        if source and (
            source.startswith("http://") or source.startswith("https://")
        ):
            url = source
        if self.__component__._pil_image:
            self._image = self.__component__._pil_image
        elif source and not url:
            self._image = PILImage.open(self.__component__.source)
        elif self.__component__.stream is not None:
            self._image = PILImage.open(self.__component__.stream)
        elif self.__component__.content:
            if isinstance(self.__component__.content, bytes):
                self._image = PILImage.open(
                    BytesIO(self.__component__.content)
                )
            elif isinstance(self.__component__.content, str):
                self._image = PILImage.open(
                    BytesIO(base64.b64decode(self.__component__.content))
                )
        elif url is not None:
            import httpx

            with httpx.Client() as client:
                response = client.get(url)
                response.raise_for_status()
                self._image = PILImage.open(BytesIO(response.content))
        else:
            raise BadRequestError("Image not initialized.")
        self._init = True

    def get_info(self) -> ImageInfo:
        self.__setup__()
        return ImageInfo(width=self._image.width, height=self._image.height)

    def convert(self, type: str, format: str | None = None) -> Any:
        self.__setup__()
        if format is not None:
            format = format.upper()
        if type == "pil":
            return self._image
        elif type == "bytes":
            if format in self._image_bytes:
                return self._image_bytes[format]
            byte_io = BytesIO()
            self._image.save(
                byte_io, format=format or self._image.format or "PNG"
            )
            self._image_bytes[format] = byte_io.getvalue()
            return self._image_bytes[format]
        elif type == "base64":
            if format in self._image_base64:
                return self._image_base64[format]
            byte_io = BytesIO()
            self._image.save(
                byte_io, format=format or self._image.format or "PNG"
            )
            self._image_base64[format] = base64.b64encode(
                byte_io.getvalue()
            ).decode("utf-8")
            return self._image_base64[format]
        raise BadRequestError(f"Type {type} not supported")

    def save(self, path: str, format: str | None = None) -> None:
        if format is not None:
            format = format.upper()
        self._image.save(path, format=format)

    def show(self):
        self._image.show()



from enum import Enum
from typing import Literal

from x8.core import Component, DataModel


class APIInfo(DataModel):
    """API Info.

    Attributes:
        host: Host address.
        port: Host port.
        ssl: SSL enabled.
    """

    host: str
    port: int | None
    ssl: bool = False


class APIAuthType(str, Enum):
    """API Auth Type Enum.

    Attributes:
        API_KEY: API Key authentication.
        BEARER_TOKEN: Bearer token (JWT) authentication.
        BASIC: Basic authentication.
    """

    API_KEY = "api_key"
    BEARER_TOKEN = "bearer_token"
    BASIC = "basic"


class APIKeyAuthConfig(DataModel):
    """API Key Auth Config.

    Attributes:
        header: Header name for API key.
        query: Query parameter name for API key.
    """

    header: str | None = None
    query: str | None = None


class APIAuth(DataModel):
    """Auth Mapping Info.

    Attributes:
        type: Auth type.
        component: Auth component.
        config: Additional config for auth.
    """

    type: APIAuthType
    component: Component | None = None
    config: APIKeyAuthConfig | None = None


class ArgSourceType(str, Enum):
    """Arg Source Type Enum.

    Attributes:
        BODY: Field from request body.
        QUERY: Field from query parameters.
        HEADER: Field from request headers.
        PATH: Field from URL path.
        AUTH: Field from auth validation.
        AUTH_USER: Field from authenticated user info.
    """

    BODY = "body"
    QUERY = "query"
    HEADER = "header"
    PATH = "path"
    AUTH = "auth"
    AUTH_USER = "auth_user"


class ArgSource(DataModel):
    """Argument Source.

    Attributes:
        type: Source type.
        field: Field name for the source.
    """

    type: ArgSourceType
    field: str | None = None


class ArgMapping(DataModel):
    """Argument Mapping Info.

    Attributes:
        name:
            Argument name.
        source:
            Argument source.
    """

    name: str
    source: ArgSource


class ResponseMapping(DataModel):
    """Response Mapping Info.

    Attributes:
        body: Body mapping field.
        headers: Headers field mapping.
        media_type: Media type.
    """

    body: str | dict[str, str] | None = None
    headers: dict[str, str] | None = None
    media_type: str | None = None


class OperationMapping(DataModel):
    """Operation Mapping Info.

    Attributes:
        name: Operation name.
        path: URL path.
        method: Http method.
        status: Status code.
        args: Arg mappings.
        auth: Auth component.
        supress: Suppress operation in API.
    """

    name: str
    path: str | None = None
    method: str | None = None
    status: int | None = None
    response: ResponseMapping | None = None
    args: list[ArgMapping] = []
    auth: APIAuth | None = None
    supress: bool = False


class ComponentMapping(DataModel):
    """Component Mapping Info.

    Attributes:
        component: Component.
        prefix: URL prefix.
        tags: Tags for the component.
        args: Common arg mappings across operations.
        auth: Auth component.
        sync_async_resolution: If both sync and async methods
            are available (with async methods prefixed with 'a'),
            how to resolve them. Can be "async", "sync", or "both".
        remove_async_prefix: If True, removes 'a' prefix from async methods
            when naming operations.
        predict_http_method: If True, predicts HTTP method
            based on operation name.
        operations: Operation mappings.
    """

    component: Component
    prefix: str | None = None
    tags: list[str] | str | None = None
    args: list[ArgMapping] = []
    auth: APIAuth | None = None
    sync_async_resolution: Literal["async", "sync", "both"] = "async"
    remove_async_prefix: bool = True
    predict_http_method: bool = False
    operations: list[OperationMapping] = []



from x8.content.audio import Audio
from x8.content.image import Image
from x8.core import DataModel


class VideoData(DataModel):
    source: str | None = None
    """Source of the video."""

    content: bytes | str | None = None
    """Content of the video."""

    media_type: str | None = None
    """Media type of the video."""


class VideoFrame(DataModel):
    timestamp: float
    """Timestamp of the frame in seconds."""

    frame_index: int | None = None
    """Index of the frame."""

    key_frame: bool
    """A value indicating whether the frame is a key frame."""

    width: int
    """Width of the video frame in pixels."""

    height: int
    """Height of the video frame in pixels."""

    image: Image
    """Image."""


class VideoAudio(DataModel):
    start: float | None = None
    """Start timestamp of the audio."""

    end: float | None = None
    """End timestamp of the audio."""

    format: str | None = None
    """Format of the audio."""

    codec: str | None = None
    """Codec of the audio"""

    audio: Audio
    """Audio."""


class VideoInfo(DataModel):
    duration: float
    """Duration of the video in seconds."""

    width: int
    """Width of the video in pixels."""

    height: int
    """Height of the video in pixels."""

    total_frames: int
    """Total number of frames in the video stream."""

    frame_rate: float
    """Frame rate of the video."""

    bit_rate: int
    """ Bit rate of the video in bits per second."""

    format: str | None = None
    """Format of the video."""

    codec: str | None = None
    """Name of the codec used for the video stream."""

    pixel_format: str | None = None
    """Pixel format of the video stream."""

    start_time: float | None = None
    """Start time of the video in seconds."""
